/*
VR2 Control
===========

Controlling the VR2 drive unit (Permobil C300 wheel chair) with an Arduino.

This sketch emulates the serial protocol used by the joystick unit of a Permobil C300 electric wheel chair.
It uses a software serial 38400 baud connection with RX and TX over the same wire - this poses some difficulties (pullups are bad here!).

The communication is very time-critical. That is why I am using a modified "CustomSoftwareSerial" library.
I needed to alter some values to make the timings perfectly. Also, I could not get it to RX correctly,
so I disabled its RX functionality (free up interrupts) and added the standard SoftwareSerial library for RX.
(It may not be able to TX at 38400,E,8,2, but when RXing it just ignores parity and superfluous stop bits - so that works).

The functionality of the joystick unit is exposed to the serial port using pybridge.h or simple terminal commands.
pybridge allows for a (theoretically) reliable transmission channel including flow control, checksums and re-transmitting.

Goal of this sketch is to control the VR2 from a Raspberry Pi running the client "arduino.py"

Robots!


2018-03-20 Bernhard "HotKey" Slawik

*/


// Chose your serial interface mode (Only one!)
#define INTERFACE_PYBRIDGE  // Accept commands through pybridge commands
//#define INTERFACE_TERMINAL  // Accept single-letter characters
//#define DEBUG // Output lots of text (uses lots of mem and interferes with binary protocols like pybridge)

#define USE_GPS_COMPASS // Include GpsCompass library?
#define USE_GPS // Include serial GPS functionality via NeoICSerial and TinyGPS
#define USE_COMPASS // Include compass functionality using HMC5883

#define USE_NEOPIXELS // Use neopixel?
#define USE_LAMP

//#define USE_JOYSTICK_ANALOG
//#define USE_JOYSTICK_WII_NUNCHUCK
#define USE_JOYSTICK_WII_CLASSIC_CONTROLLER

#define SERIAL_BAUD 57600
//#define SERIAL_BAUD 115200  // Seems to have minor issues

#define LED_PIN 13


#ifdef USE_JOYSTICK_ANALOG
  #define JOY_X_ANALOG_PIN A0
  #define JOY_Y_ANALOG_PIN A1
  #define JOY_BUTTON_PIN 7
  #define JOY_X_CENTER 512
  #define JOY_Y_CENTER 512
  bool joystick_debouncing = true;  // Start de-bouncing. Why not?
#endif

#ifdef USE_JOYSTICK_WII_NUNCHUCK
  #include "joystick_wn.h"
  bool joystick_button_z_last = false;
#endif
#ifdef USE_JOYSTICK_WII_CLASSIC_CONTROLLER
  #include "joystick_wcc.h"
#endif

bool joystick_enabled = false;
int joy_x = 0;
int joy_y = 0;


#define JOYSTICK_SCHEME_2D 0
#define JOYSTICK_SCHEME_TANK 1
#define JOYSTIC_SCHEMES 2
int joystick_scheme = JOYSTICK_SCHEME_2D;


#ifdef USE_LAMP
  #define LAMP_PIN 5  // 4 or 5
  bool lamp_on = false;
#endif
bool flash = false;
bool blackout = false;


#ifdef DEBUG
  #define put(t) Serial.println(t)
  #define put_(t) Serial.print(t)
  #define put_hex_(t) Serial.print(t, HEX)
  #define put_float_(t) Serial.print(t, 6)
#else
  #define put(t) ;
  #define put_(t) ;
  #define put_hex_(t) ;
  #define put_dump(data, l) ;
  #define put_float_(t) ;
#endif






////////// VR2
//#define VR2_DEBUG // Verbosely log messages. This will interfere with binary protocols (such as pybridge)!
//#define VR2_AUTO_POWER_ON // Automatically init?
#include "vr2.h"




////////// NeoPixel
#ifdef USE_NEOPIXELS
  /*
  index description
        Head R
        Head L
        Tail R
        Tail L
        Ring (24)
  */
  
  #define PIXELS (1+1+1+1 + 24)
  
  // Arduino UNO Pin 12: Port D, bit 4, DDRB
  #define PIXEL_PORT  PORTB
  #define PIXEL_DDR   DDRB
  #define PIXEL_BIT   4
  #include "neopixel.h"

  
  #define NEOPIXEL_UPDATE_MILLIS 50
  #define NEOPIXEL_AUTO_IDLE 10000  // Timeout after CMD_NEOPIXEL_xxx after which we continue with auto animation
  #define NEOPIXEL_BLINK_INTERVAL 300
  #define NEOPIXEL_BLINK_THRESHOLD 20 // Threshold for automatic indicator lights

  #define COLOR_HEAD_FULL 0xf0f0f0 // Head light RGB
  #define COLOR_HEAD_DIM 0x202020 // Head light RGB
  
  #define COLOR_TAIL_FULL 0x400000 // Tail light RGB
  #define COLOR_TAIL_DIM 0x200000 // Tail light RGB
  
  #define COLOR_INDICATOR 0xd06000  // Indicator light RGB
  #define COLOR_INDICATOR_DIM 0x0a0400  // Indicator light RGB dimmed
  
  #define COLOR_FLASH 0x808080	//0xf0f0f0  // Flash (not too high, since ALL LEDs will flash!)
  #define COLOR_IDLE 0x040000 // Default idle color

  #define COLOR_LAMP 0x989080 // Color of stand lamp
  #define COLOR_LAMP_IDLE 0x100000 // Color of stand lamp chase
  
  
  unsigned long timeNeopixelLast;
  
  unsigned int neopixel_auto_frame = 0;
  void neopixel_auto_draw() {

    byte cR, cG, cB;
    unsigned long c;
    
    unsigned int f;
    unsigned int i;

    bool blinkState = (neopixel_auto_frame / (NEOPIXEL_BLINK_INTERVAL / NEOPIXEL_UPDATE_MILLIS)) % 2 == 0;

    bool indicatorLeft = (joy_x >= NEOPIXEL_BLINK_THRESHOLD);
    bool indicatorRight = (joy_x <= -NEOPIXEL_BLINK_THRESHOLD);
    bool indicatorActive = (indicatorLeft || indicatorRight);
    bool reverse = (joy_y <= -NEOPIXEL_BLINK_THRESHOLD);

    
    cli();
    if (blackout) {
      for(i = 0; i < 4; i++) { // Send all to black
        neopixel_sendPixel2(0x000000);
      }
    } else
    if (flash) {

      //@TODO: Police light!!!
      
      for(i = 0; i < 4; i++) {
        neopixel_sendPixel2(COLOR_FLASH);
      }
    } else
    if (isConnected) {
      // Connected

      // 0: Head R
      if (indicatorRight && blinkState) neopixel_sendPixel2(COLOR_INDICATOR);
      else if (joystick_enabled) {
        if (reverse)                    neopixel_sendPixel2(COLOR_HEAD_DIM);
        else                            neopixel_sendPixel2(COLOR_HEAD_FULL);
      }
      else                              neopixel_sendPixel2(COLOR_HEAD_DIM);
      
      // 1: Head L
      if (indicatorLeft && blinkState)  neopixel_sendPixel2(COLOR_INDICATOR);
      else if (joystick_enabled) {
        if (reverse)                    neopixel_sendPixel2(COLOR_HEAD_DIM);
        else                            neopixel_sendPixel2(COLOR_HEAD_FULL);
      }
      else                              neopixel_sendPixel2(COLOR_HEAD_DIM);
      
      // 2: Tail R
      if (indicatorRight && blinkState) neopixel_sendPixel2(COLOR_INDICATOR);
      else if (joystick_enabled) {
        if (reverse)                    neopixel_sendPixel2(COLOR_TAIL_DIM);
        else                            neopixel_sendPixel2(COLOR_TAIL_FULL);
      }
      else                              neopixel_sendPixel2(COLOR_TAIL_DIM);
      
      // 3: Tail L
      if (indicatorLeft && blinkState)  neopixel_sendPixel2(COLOR_INDICATOR);
      else if (joystick_enabled) {
        if (reverse)                    neopixel_sendPixel2(COLOR_HEAD_FULL);
        else                            neopixel_sendPixel2(COLOR_TAIL_FULL);
      }
      else                              neopixel_sendPixel2(COLOR_TAIL_DIM);
      
      
    } else {
      // Not connected to VR2
      
      /*
      // Yellow warn blink on all LEDs
      if (blinkState) { cR = 0x80;  cG = 0x60; cB = 0x00; }
      else            { cR = 0x00;  cG = 0x00; cB = 0x00; }
      
      */

      if (reconnect) {
        // Trying to connect: Blink
        if (blinkState) c = COLOR_INDICATOR;  //{ cR = 0x80;  cG = 0x60; cB = 0x00; }
        else            c = COLOR_INDICATOR_DIM;  //{ cR = 0x10;  cG = 0x0c; cB = 0x00; }
      }
      else              c = COLOR_IDLE;       //{ cR = 0x00;  cG = 0x04; cB = 0x00; }

      for(i = 0; i < 4; i++) {
        //neopixel_sendPixel(cR, cG, cB);
        neopixel_sendPixel2(c);
      }
      
    }


    // Rest of the LEDs
    
    if (blackout) {
      for(i = 4; i < PIXELS; i++) {
        neopixel_sendPixel(0x00, 0x00, 0x00);
      }
      
    }
    
    #ifdef USE_LAMP
    else if (lamp_on) {
      for(i = 4; i < PIXELS; i++) {
        neopixel_sendPixel(COLOR_LAMP);
      }
    }
    #endif
    else {
      c = COLOR_LAMP_IDLE;
      
      // Chase using the given color
      f = (neopixel_auto_frame % 4);
      for(i = 4; i < PIXELS; i++) {
        if ((i % 8) == f)
          neopixel_sendPixel(c);
        else
          neopixel_sendPixel(0x00, 0x00, 0x00);
      }
    }
    
    
    sei();
    neopixel_show();
    
    //neopixel_auto_highlight = (neopixel_auto_highlight + 1);  // % PIXELS;
    neopixel_auto_frame++;
  }

  void neopixel_showStatus(byte s) {
    #define NEOPIXEL_MAX_STATUS 5
    //int si = (s * PIXELS) / NEOPIXEL_MAX_STATUS;
    
    cli();
    for(unsigned int i=0; i < PIXELS; i++) {
      if (i < s)
        neopixel_sendPixel(0x20, 0x20, 0x00);
      else
      if (i <= NEOPIXEL_MAX_STATUS)
        neopixel_sendPixel(0x01, 0x01, 0x00);
      else
        neopixel_sendPixel(0x01, 0x01, 0x01);
    }
    sei();
    neopixel_show();
  }

  
#endif




////////// pybridge stuff
#ifdef INTERFACE_PYBRIDGE
  // Configure pybridge (defines before including the h-file):
  // Use standard Serial for communication and debugging
  #define SerialCommand Serial
  #define SerialDebug Serial
  //#define DEBUG_VERBOSE
  
  // Select functionality
  //#define USE_DIGITALS
  //#define USE_ANALOGS
  //#define USE_SERVOS
  //#define USE_LEDRGB
  //#define LEDRGB_NUM 8+9+8+9
  //#define LEDRGB_PIN 6
  
  #define USE_CUSTOM_COMMANDS // Enable pybridge_on_command callback
  
  #include "pybridge.h"
  
  #define PYB_STATE_INTERVAL_MILLS 500  // Milliseconds between state reports
  
  // Custom pybridge commands
  #define CMD_VR2_INFO 'i'  // Used to report small infos
  #define CMD_VR2_STATE 's' // Used to report current system state
  #define CMD_VR2_POWER_OFF 0
  #define CMD_VR2_POWER_ON 1
  #define CMD_VR2_JOY 'j'
  #define CMD_VR2_KEY 'k'
  #define CMD_VR2_STOP '.'
  //#define CMD_VR2_FASTER '+'
  //#define CMD_VR2_SLOWER '-'
  
  #define CMD_LOCATION 'L'
  #define CMD_MAGNET 'M'
  //#define CMD_HEADING 'H'
  #define CMD_NEOPIXEL_FILL 'P'
  #define CMD_NEOPIXEL_GRADIENT 'G'
  #define CMD_NEOPIXEL_RAW 'W'
  
  #define INFO_INIT 0x00
  #define INFO_READY 0x01
  #define INFO_DISCONNECTED 0x02
  #define INFO_ERROR 0x03
  //#define INFO_GPS 0x04
  
  
  void pyb_send_info(byte i) {
    // Report simple status byte
    sendCommand_byte(CMD_VR2_INFO, i);
  }
  
  void pyb_send_state() {
    // All interesting state vars in one packet
    byte data[7] = {isConnected, state_flags1, state_flags2, state_speed, state_battery, (uint8_t)joyX, (uint8_t)joyY};
    sendCommand_data(CMD_VR2_STATE, 7, data);
    
    // Make sure we don't flood the line afterwards
    timePybStateLast = millis();
  }
  
  
  byte pyb_on_command(byte cmd, byte len, byte *data) {
    //put(F("pyb_on_command"));
    byte x;
    byte o;
    byte i;
    
    
    switch(cmd) {
      case CMD_VR2_POWER_ON:
        reconnect = 1;
        vr2_init();
        return 1;
  
      case CMD_VR2_POWER_OFF:
        reconnect = 0;
        vr2_power_off();
        return 1;
        
      case CMD_VR2_JOY:
        if (len < 2) return 1; // Stop propagation
        
        put_(F("CMD_VR2_JOY x=")); put_((int8_t)data[0]); put_(F(" y=")); put((int8_t)data[1]);
        vr2_set_joystick((int8_t)data[0], (int8_t)data[1]);
        
        return 1; // Stop propagation
  
      case CMD_VR2_KEY:
        if (len < 1) return 1; // Stop propagation
        
        put_(F("CMD_VR2_KEY k=")); put((int8_t)data[0]);
        vr2_press_key((int8_t)data[0]);
        
        return 1; // Stop propagation
  
      case CMD_VR2_STOP:
        vr2_set_joystick(0, 1);
        return 1;
      
      /*
      case CMD_VR2_FASTER:
        vr2_press_key(VR2_KEY_FASTER);
        return 1;
        
      case CMD_VR2_SLOWER:
        vr2_press_key(VR2_KEY_FASTER);
        return 1;
      */
      
      case CMD_NEOPIXEL_FILL:
        #ifdef USE_NEOPIXELS
          if (len == 3) {
            neopixel_showColor((int8_t)data[0], (int8_t)data[1], (int8_t)data[2]);
          }
          
          if (len == 4) {
            
            o = data[3];
            cli();
            for(x = 0; x < o; x++) {
              neopixel_sendPixel((int8_t)data[0], (int8_t)data[1], (int8_t)data[2]);
            }
            sei();
            neopixel_show();
          }
        #endif
        //neopixel_auto = false;
        timeNeopixelLast = millis() + NEOPIXEL_AUTO_IDLE;
        return 1;
      
      case CMD_NEOPIXEL_RAW:
        #ifdef USE_NEOPIXELS
          x = 0;
          o = 0;
          cli();
          while ((o < len) && (x < PIXELS)) {
            neopixel_sendPixel(data[o++], data[o++], data[o++]);
            x++;
          }
          sei();
          neopixel_show();
        #endif
        //neopixel_auto = false;
        timeNeopixelLast = millis() + NEOPIXEL_AUTO_IDLE;
        return 1;
      
      case CMD_NEOPIXEL_GRADIENT:
        #ifdef USE_NEOPIXELS
          
          byte gr,gg,gb;
          byte nr,ng,nb = 0;
          byte runlength = 0;
          x = 0;
          o = 0;
          cli();
          while ((o < len) && (x < PIXELS)) {
            // Constant color (initially black)
            gr = nr;
            gg = ng;
            gb = nb;
            runlength = data[o++];
            for (i = 0; i < runlength; i++) {
              neopixel_sendPixel(gr,gg,gb);
              x++;
            }
            
            // New color
            nr = data[o++];
            ng = data[o++];
            nb = data[o++];
            
            // Gradient length
            runlength = data[o++];
            /*
            byte r = runlength+1;
            for (i = 0; i < runlength; i++) {
              byte j = i+1;
              neopixel_sendPixel(
                gr + (int16_t)((nr - gr) * j) / r,
                gg + (int16_t)((ng - gg) * j) / r,
                gb + (int16_t)((nb - gb) * j) / r
              );
              x++;
              if (x >= PIXELS) break;
            }
            */
            byte r = runlength+1;
            int16_t dr = nr - gr;
            int16_t dg = ng - gg;
            int16_t db = nb - gb;
            dr /= r;
            dg /= r;
            db /= r;
            for (i = 0; i < runlength; i++) {
              gr += dr;
              gg += dg;
              gb += db;
              neopixel_sendPixel(gr, gg, gb);
              x++;
              //if (x >= PIXELS) break;
            }
            
          }
          sei();
          neopixel_show();
        #endif
        //neopixel_auto = false;
        timeNeopixelLast = millis() + NEOPIXEL_AUTO_IDLE;
        return 1;
    }
    
    return 0; // 0=Continue, 1=Stop propagation
  }


#endif


////////// GpsCompass
#ifdef USE_GPS_COMPASS
  #include "gpsCompass.h"
  
  #ifdef USE_GPS
    uint32_t deg_to_uint32(float d) {
      float f = d;
      while (f < 0.0f) f = f + 360.0f;
      f = f / 360.0f;
      f = f * (float)(0x100000000L); 
      return f;
    }
    uint16_t deg_to_uint16(float d) {
      float f = d;
      while (f < 0.0f) f = f - 360.0f;
      return (float)(0x10000) * (f / 360.0f);
    }

    void on_gpsCompass_position(float lat, float lon, float alt) {
    
      #ifdef INTERFACE_PYBRIDGE
        // Report GPS info via pybridge
    
        // lat: [-90...90]
        // lon: [-180...180]
        // alt: [-7000...inf]
        uint8_t data[4+4+2];
    
        uint32_t u_lat = deg_to_uint32(lat);
        data[0] = (uint8_t)(u_lat >> 24);
        data[1] = (uint8_t)(u_lat >> 16);
        data[2] = (uint8_t)(u_lat >>  8);
        data[3] = (uint8_t)(u_lat      );
    
        uint32_t u_lon = deg_to_uint32(lon);
        data[4] = (uint8_t)(u_lon >> 24);
        data[5] = (uint8_t)(u_lon >> 16);
        data[6] = (uint8_t)(u_lon >>  8);
        data[7] = (uint8_t)(u_lon      );

        uint16_t u_alt = (uint16_t)(alt + 32768.0f);
        data[8] = (uint8_t)(u_alt >>  8);
        data[9] = (uint8_t)(u_alt      );
    
        sendCommand_data(CMD_LOCATION, 4+4+2, &data[0]);
    
        //pyb_send_info(INFO_GPS);
      #endif
      
      #ifdef INTERFACE_TERMINAL
        put_(F("Position: "));
        put_float_(lat);
        put_(F(", "));
        put_float_(lon);
        
        put_("\tAltitude: ");
        put_(alt);
        
        put_("\tDate: ");
        put_(gpsCompass_time_year);
        put_('-');
        put_(gpsCompass_time_month);
        put_('-');
        put_(gpsCompass_time_day);
      
        put_("\tTime: ");
        put(gpsCompass_time_millis);  
      #endif
    }
  #endif
  
  #ifdef USE_COMPASS
    #define MAX_MAGNET_UT 1000.0f // Biggest/smallest value (+-) to encode (in microtesla)
    uint16_t uT_to_uint16(float t) {
      // -1000 < t < 1000
      float d = (t / MAX_MAGNET_UT);
      if (d < -1.0f) d = -1.0f;
      else if (d > 1.0f) d = 1.0f;
      return (float)(0xffff) * ((d + 1.0f) / 2.0f);
    }
    void on_gpsCompass_magnet(float x, float y, float z) {
    
      #ifdef INTERFACE_PYBRIDGE
        // Report GPS info via pybridge
    
        uint16_t sx = uT_to_uint16(x);
        uint16_t sy = uT_to_uint16(y);
        uint16_t sz = uT_to_uint16(z);
    
        uint8_t data[3*2];
        data[0] = (uint8_t)((sx >> 8 ) & 0xff);
        data[1] = (uint8_t)((sx      ) & 0xff);
        data[2] = (uint8_t)((sy >> 8 ) & 0xff);
        data[3] = (uint8_t)((sy      ) & 0xff);
        data[4] = (uint8_t)((sz >> 8 ) & 0xff);
        data[5] = (uint8_t)((sz      ) & 0xff);
        sendCommand_data(CMD_MAGNET, 3*2, &data[0]);
        //pyb_send_info(INFO_GPS);
          
      #endif
      
      #ifdef INTERFACE_TERMINAL
        put_("Magnet: ");
        put_float_(x); put_(", "); put_float_(y); put_(", "); put_float_(z); put("");
      #endif
      
    }
    
    /*
    void on_gpsCompass_heading() {
    
      #ifdef INTERFACE_PYBRIDGE
        // Report GPS info via pybridge
    
        uint16_t heading = deg_to_uint16(gpsCompass_heading_deg);
    
        uint8_t data[2];
        data[0] = (uint8_t)((heading >> 8 ) & 0xff);
        data[1] = (uint8_t)((heading      ) & 0xff);
        sendCommand_data(CMD_HEADING, 2, &data[0]);
        
        //pyb_send_info(INFO_GPS);
          
      #endif
      
      #ifdef INTERFACE_TERMINAL
        put_("\tHeading: ");
        put_(gpsCompass_heading_deg);
      #endif
      
    }
    */
  #endif
#endif


//////////////////// Main Implementation ////////////////////

////////// VR2 callbacks
void on_vr2_init() {
  // Init loop has been started!
  #ifdef INTERFACE_PYBRIDGE
    pyb_send_info(INFO_INIT);
  
    // Report state to host immediately
    pyb_send_state();
    
  #endif
}

void on_vr2_disconnect() {
  #ifdef INTERFACE_PYBRIDGE
    pyb_send_info(INFO_DISCONNECTED);
    
    // Report state immediately
    pyb_send_state();
  #endif
}

void on_vr2_ready() {
  #ifdef INTERFACE_PYBRIDGE
    pyb_send_info(INFO_READY);
  #endif
}

void on_vr2_state() {
  // VR2 state has changed
  
  #ifdef INTERFACE_PYBRIDGE
    // Send through pybridge
    // Force a new status report ASAP by setting the last time to past
    timePybStateLast = 0;
  #endif
}

void on_vr2_keepAlive() {
  // Called during longer processes to keep other things alive
  
  #ifdef USE_GPS_COMPASS
    // handle GpsCompass regularly or the serial buffer will overflow!
    gpsCompass_loop();
  #endif
}


void setup() {
  
  // Open serial communications and wait for port to open:
  Serial.begin(SERIAL_BAUD);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }
  
  //put(F("VR2 Control"));
  
  pinMode(LED_PIN, OUTPUT);
  digitalWrite(LED_PIN, HIGH);

  #ifdef USE_LAMP
    pinMode(LAMP_PIN, INPUT_PULLUP);
    lamp_on = (digitalRead(LAMP_PIN) == LOW);
  #endif
  
  #ifdef USE_NEOPIXELS
    // Neopixel setup
    neopixel_setup();
    
    //neopixel_showColor(0x10, 0x00, 0x00);
    neopixel_showStatus(0);
    delay(100);
    
    timeNeopixelLast = millis();
  #endif
  
  #ifdef USE_JOYSTICK_ANALOG
    // Setup joystick
    pinMode(JOY_BUTTON_PIN, INPUT_PULLUP);
  #endif
  #ifdef USE_JOYSTICK_WII_NUNCHUCK
    nunchuck_setup(true);
  #endif
  #ifdef USE_JOYSTICK_WII_CLASSIC_CONTROLLER
    joystick_wcc_setup(true);
  #endif
  
  #ifdef USE_NEOPIXELS
    neopixel_showStatus(1);
    delay(100);
  #endif

  
  
  #ifdef INTERFACE_PYBRIDGE
    // Setup pybridge...
    pybridge_on_command = pyb_on_command; // Catch our own commands
    pybridge_setup();
  #endif

  #ifdef USE_NEOPIXELS
    //neopixel_showColor(0x10, 0x08, 0x00);
    neopixel_showStatus(2);
    delay(100);
  #endif

  
  // Main VR2 setup
  vr2_on_state = on_vr2_state;  // Set callback for state changes
  vr2_on_init = on_vr2_init;
  vr2_on_keepAlive = on_vr2_keepAlive;
  vr2_on_disconnect = on_vr2_disconnect;
  vr2_on_ready = on_vr2_ready;
  vr2_setup();
  
  #ifdef USE_NEOPIXELS
    //neopixel_showColor(0x10, 0x10, 0x00);
    neopixel_showStatus(3);
    delay(100);
  #endif

  
  // Setup GPS and compass
  #ifdef USE_GPS_COMPASS

    #ifdef USE_GPS
      gpsCompass_on_position = on_gpsCompass_position;
    #endif
    #ifdef USE_COMPASS
      gpsCompass_on_magnet = on_gpsCompass_magnet;
    #endif

    gpsCompass_setup(false);

    #ifdef USE_NEOPIXELS
      //neopixel_showColor(0x00, 0x10, 0x00);
      neopixel_showStatus(4);
      delay(100);
    #endif
    
    // Initial handling
    for(int i = 0; i < 10; i++) {
      gpsCompass_loop();
      delay(1);
    }
    
  #endif
  #ifdef USE_NEOPIXELS
    //neopixel_showColor(0x00, 0x10, 0x00);
    neopixel_showStatus(5);
    delay(100);
  #endif

  
  
  #ifdef INTERFACE_PYBRIDGE
    // Signal ready
    sendCommand_none(CMD_READY);
  #endif
  
  
  digitalWrite(LED_PIN, LOW);
  
  #ifdef USE_NEOPIXELS
    //neopixel_showColor(0x00, 0x10, 0x00);
    neopixel_showStatus(6);
  #endif

}



void loop() { // run over and over
  unsigned long timeNow = millis();
  int timeDelta;
  
  // Update VR2 connection
  vr2_loop();
  
  
  // pybridge stuff
  #ifdef INTERFACE_PYBRIDGE
    
    // Send regular state to host
    timeDelta = timeNow - timePybStateLast;
    if ((timeDelta > PYB_STATE_INTERVAL_MILLS) || (timeDelta < 0)) {
      
      // Report state to host
      pyb_send_state();
      
    }
    
    // Check for pybridge commands
    pybridge_loop();
    
  #endif

  // handle GpsCompass
  #ifdef USE_GPS_COMPASS
    gpsCompass_loop();
  #endif

  #ifdef USE_LAMP
    lamp_on = (digitalRead(LAMP_PIN) == LOW);
  #endif
  flash = false;
  blackout = false;

  
  #ifdef USE_JOYSTICK_ANALOG
    // handle manual control through analog joystick
    bool joystick_button = (digitalRead(JOY_BUTTON_PIN) == LOW);

    /*
    Serial.println();
    //Serial.print(joy_x);
    //Serial.print("\t");
    //Serial.print(joy_y);
    Serial.print(joystick_button);
    Serial.println();
    */
    
    if (joystick_debouncing) {
      // De-bouncing mode (ignore everything until button released)
      
      if (joystick_button == false) joystick_debouncing = false;
      
    } else {
      // Regular mode (handle input)
      
      if (joystick_button) {
        // Toggle manual control
        joystick_enabled = !joystick_enabled;
        joystick_debouncing = true; // Need to de-bounce
        delay(1);
        
      } else {
        // Nothing pressed
        if (joystick_enabled) {
          //Joy control!

          // Make sure we are connected
          if (isConnected == 0)
            reconnect = 1;
        
          joy_x = map(analogRead(JOY_X_ANALOG_PIN) - JOY_X_CENTER, -512, 512, -99, 99);
          joy_y = map(analogRead(JOY_Y_ANALOG_PIN) - JOY_Y_CENTER, -512, 512, -99, 99);

           // Set joystick position
          vr2_set_joystick((int8_t)joy_x, (int8_t)joy_y);
        }
      }
      
    }
  #endif
  
  #ifdef USE_JOYSTICK_WII_NUNCHUCK
    nunchuck_loop();
    boolean joystick_button_z = ((nunchuck_buttons & 0x01) > 0);
    boolean joystick_button_b = ((nunchuck_buttons & 0x02) > 0);
    /*
    Serial.println();
    Serial.print(nunchuck_buttons);
    Serial.print("\t");
    Serial.print(nunchuck_joy_x);
    Serial.print("\t");
    Serial.print(nunchuck_joy_y);
    Serial.println();
    */
    
    if (joystick_button_b) {
      // E-STOP!
      vr2_set_joystick(0,0);
      
      if (isConnected) {
        // Disconnect
        vr2_power_off();
      }
      reconnect = 0;
      
    } else {
      if (joystick_button_z) {
        // Z is pressed
        
        // Make sure we are connected
        if (isConnected == 0) {
          reconnect = 1;
        }
        
        joy_x = constrain(-nunchuck_joy_x, -99, 99);
        joy_y = constrain(-nunchuck_joy_y, -99, 99);
        vr2_set_joystick((int8_t)(joy_x), (int8_t)(joy_y));
  
        
      } else {

        /*
        // Z is released
        if (joystick_button_z_last) {
          // Button was just released.
  
          // Reset to 0
          vr2_set_joystick(0,0);
          
        }
        */
      }
      
    }
    joystick_button_z_last = joystick_button_z;
  #endif
  #ifdef USE_JOYSTICK_WII_CLASSIC_CONTROLLER
  
    joystick_wcc_loop();

    // Home button (re)connects to the VR2
    if (wcc_state.ButtonHome) {
      reconnect = true;
    }

    // ZL/ZR areemergency stop
    if ((wcc_state.ButtonZL) || (wcc_state.ButtonZR)) {
      // E-STOP(s)

      vr2_set_joystick(0,0);
      
      if (isConnected) {
        // Disconnect
        vr2_power_off();
      }
      reconnect = 0;
    }

    // + (start) and - (select) shift "gears"
    if (wcc_state.ButtonStart) {
      vr2_press_key(VR2_KEY_FASTER);
    }
    if (wcc_state.ButtonSelect) {
      vr2_press_key(VR2_KEY_SLOWER);
    }

    
    if (wcc_state.ButtonA) {
      // Honk
      vr2_press_key(VR2_KEY_HORN);
      
      lamp_on = true; // Force lamp on
      flash = true;
    }
    if (wcc_state.ButtonB) {
      blackout = true;
    }

    // Switch joystic schemes
    if (wcc_state.ButtonX) {
      joystick_scheme = JOYSTICK_SCHEME_2D;
    }
    if (wcc_state.ButtonY) {
      joystick_scheme = JOYSTICK_SCHEME_TANK;
    }

    // L/R shoulder buttons are the "dead man switch"
    if ((wcc_state.ButtonL) || (wcc_state.ButtonR)) {
      // Joystick enabled
      
      if (!joystick_enabled) {
        // Just activated manual joystick control
        //@TODO: Flash LEDs or something
        
      } else {
        // Still active
        
      }
      joystick_enabled = true;
      
    } else {
      
      // Joystick disabled
      
      if (joystick_enabled) {
        // The dead man switch has JUST been released!
        
        // Restore sane standby state:
        vr2_set_joystick(0,0);  // Return joystick to neutral (Important!!!!!)
      }
      
      joystick_enabled = false;
    }
    

    
    if (joystick_enabled) {
      
      // Convert Wii joystick pos to percentages
      int joy_left_x = (int)wcc_state.LeftX * -6;
      int joy_left_y = (int)wcc_state.LeftY * 6;
      int joy_right_x = (int)wcc_state.RightX * -7;
      int joy_right_y = (int)wcc_state.RightY * 7;
      
      switch (joystick_scheme) {
        case JOYSTICK_SCHEME_2D:
          // Standard analog 2d input scheme
          joy_x = constrain(joy_left_x + joy_right_x, -99, 99);
          joy_y = constrain(joy_left_y + joy_right_y, -99, 99);
          break;
          
        case JOYSTICK_SCHEME_TANK:
          // Tank style input scheme
          joy_x = constrain(-(joy_left_y - joy_right_y) + joy_left_x + joy_right_x, -99, 99);
          joy_y = constrain((joy_left_y + joy_right_y) / 2, -99, 99);
          break;
      }

      vr2_set_joystick((int8_t)(joy_x), (int8_t)(joy_y));
    }
    
  #endif
  
  
  // Neopixel update
  #ifdef USE_NEOPIXELS
    timeDelta = timeNow - timeNeopixelLast;
    if ((timeDelta > NEOPIXEL_UPDATE_MILLIS)) { // || (timeDelta < 0)) {
      neopixel_auto_draw();
      timeNeopixelLast = timeNow;
    }
  #endif
  
  // Terminal stuff
  #ifdef INTERFACE_TERMINAL
    // Handle serial commands (interactive serial mode)
    if (Serial.available()) {
      //vr2Serial.write(Serial.read());
      char c = Serial.read();
      switch(c) {
        case 'i':
          reconnect = 1;
          vr2_init();
          break;
        
        case 'o':    // Off
          reconnect = 0;
          vr2_power_off();
          break;

        case 's':    // stop
          vr2_set_joystick(0, 1);
          break;

        case 'f':    // forward
          vr2_set_joystick(joyX, joyY+1);
          break;
        case 'b':    // backward
          vr2_set_joystick(joyX, joyY-1);
          break;
        case 'l':
          vr2_set_joystick(joyX+1, joyY);
          break;
        case 'r':
          vr2_set_joystick(joyX-1, joyY);
          break;
        
        case '+':
          vr2_press_key(VR2_KEY_FASTER);
          break;
        case '-':
          vr2_press_key(VR2_KEY_SLOWER);
          break;
          
        case 10:
        case 13:
          // CR/LF
          break;
      }
    }
  #endif
}

