#ifndef GPS_COMPASS_H
#define GPS_COMPASS_H

/*

Interfacing the ublox NEO M8N GPS and Compass module
Using "NeoICSerial" (low-footprint software serial) for the GPS
and I2C/Adafruit HMC library for the Compass


Arduino UNO connections:
  * ublox NEO uses 3.3v levels! NOT 5v tolerant (3.6v max!)
  * Vcc can be 3v or 5v, NEO has a regulator for that, but NOT for the data pins
  * 3.3v pull-ups on I2C lines (because I2C transmits by pulling line to GND)
  * 5v-to-3v voltage divider (3.3 + 2.2) on Arduino_TX-NEO_RX
  * Arduino_RX-NEO_TX can be connected directly.


This works:
  Arduino                       ublox NEO M8N
  
  5v -------------------------- Vcc (red)
  8 (RX) ---------------------- TX (green)
  9 (TX) ---3.3k----+---------- RX (yellow)
                    |
                    2.2k Ohm
                    |
  GND --------------+---------- GND (black)
  
  
  SDA (A4) ---------+---------- SDA (white)
                    |
                    3.3k Ohm (pull up)
                    |
  3.3V -------------+
                    |
                    3.3k Ohm (pull up)
                    |
  SCL (A5) ---------+---------- SCL (orange)


*/

// Debug data directly to serial? (needs to be initialized from outside)
//#define GPS_DEBUG

//#define USE_GPS // Include serial GPS functionality via NeoICSerial and TinyGPS
//#define USE_COMPASS // Include compass functionality using HMC5883


#ifdef GPS_DEBUG
  // Debugging
  #define gps_put(t) Serial.println(t)
  #define gps_put_(t) Serial.print(t)
  #define gps_put_float(t) Serial.println(t, 6)
  #define gps_put_float_(t) Serial.print(t, 6)

  #
  
#else
  // No debugging
  #define gps_put(t) ;
  #define gps_put_(t) ;
  #define gps_put_float(t) ;
  #define gps_put_float_(t) ;
#endif


#ifdef USE_GPS
  // GPS
  #include <NeoICSerial.h>

  // These pins are mandatory for you hardware platform, e.g. Arduino UNO only supports RX=8, TX=9
  #define GPS_SERIAL_RX_PIN 8
  #define GPS_SERIAL_TX_PIN 9
  #define GPS_SERIAL_BAUD 9600

  NeoICSerial gpsSerial;

  //TinyGPS++ works fine with SoftSerial, but for some reason does not report any data when used with NeoICSerial! Checksums?
  #define TINYGPS_IGNORE_CHECKSUM
  #include <TinyGPS++.h>
  TinyGPSPlus gps;

  #define TINYGPS_BUFFER_SIZE 128
  volatile char tinyGpsBuf[TINYGPS_BUFFER_SIZE];
  volatile uint8_t tinyGpsBufIn = 0;
  volatile uint8_t tinyGpsBufOut = 0;
  
  // Actual data (access them as you want, but they may get updated inside an interrupt)
  volatile float gpsCompass_position_lat = 0.0f;
  volatile float gpsCompass_position_lon = 0.0f;
  volatile float gpsCompass_position_alt = 0.0f;
  volatile uint16_t gpsCompass_time_year = 0;
  volatile uint8_t gpsCompass_time_month = 0;
  volatile uint8_t gpsCompass_time_day = 0;
  volatile unsigned long gpsCompass_time_millis = 0UL;
  
  // User callback
  typedef void (*t_gpsCompass_on_position)(float lat, float lon, float alt);
  t_gpsCompass_on_position gpsCompass_on_position = NULL;
  
  volatile bool gpsCompass_position_updated = false;
  
  
  
  void gpsCompass_handle_gps() {
    // Called whenever there is GPS data. Return as quickly as possible!
    
    //gps_put(F("gpsCompass_handle_gps!"));
    
    bool changed = false;
    
    if (gps.location.isValid()) {
      gps_put_float_(gps.location.lat()); gps_put_(F(", ")); gps_put_float(gps.location.lng());
      
      gpsCompass_position_lat = gps.location.lat();
      gpsCompass_position_lon = gps.location.lng();
      changed = true;
    }
    
    if (gps.altitude.isUpdated()) { //isValid
      gps_put_(F("Altitude (m): ")); gps_put(gps.altitude.meters());
      
      gpsCompass_position_alt = gps.altitude.meters();
      changed = true;
    }

    if (gps.date.isValid()) {
      gps_put_(F("Date: ")); gps_put_(gps.date.year()); gps_put_('-'); gps_put_(gps.date.month()); gps_put_('-'); gps_put(gps.date.day());
      
      gpsCompass_time_year = gps.date.year();
      gpsCompass_time_month = gps.date.month();
      gpsCompass_time_day = gps.date.day();
      changed = true;
    }

    if (gps.time.isValid()) {
      gps_put_(F("Time: ")); gps_put_(gps.time.hour()); gps_put_(':'); gps_put_(gps.time.minute()); gps_put_(':'); gps_put_(gps.time.second()); gps_put_('.'); gps_put(gps.time.centisecond());
      
      gpsCompass_time_millis = (unsigned long)gps.time.hour();
      gpsCompass_time_millis *= 60;
      gpsCompass_time_millis += (unsigned long)gps.time.minute();
      gpsCompass_time_millis *= 60;
      gpsCompass_time_millis += (unsigned long)gps.time.second();
      gpsCompass_time_millis *= 100;
      gpsCompass_time_millis += (unsigned long)gps.time.centisecond();
      gpsCompass_time_millis *= 10;
      changed = true;
    }

    if (changed)
      gpsCompass_position_updated = true;
  }

  // Callback for the NeoICSerial RX
  static void on_gpsSerial_rx(uint8_t c) {
    // Handle incoming byte, return as quickly as possible!

    // Debug traffic
    //gps_put_((char)c);
    /*
    if (gps.encode((char)c)) {
      gpsCompass_handle_gps();
    }
    */
    
    // Buffer data
    //if (tinyGpsBufOfs < TINYGPS_BUFFER_SIZE) {
    tinyGpsBuf[tinyGpsBufIn] = c;
    tinyGpsBufIn = (tinyGpsBufIn + 1) % TINYGPS_BUFFER_SIZE;
    
  }
#endif


#ifdef USE_COMPASS
  // Compass
  #include <Wire.h>
  /*
  //// Using Adafruit
  #include <Adafruit_Sensor.h>
  #include <Adafruit_HMC5883_U.h>
  // Assign a Unique ID to the HMC5883 Compass Sensor
  Adafruit_HMC5883_Unified gpsMag = Adafruit_HMC5883_Unified(12345);
  bool gpsCompass_magnet_found = false;
  */
  
  //// Manual I2C as in https://www.sparkfun.com/tutorials/301
  #define HMC5883_ADDRESS 0x1E
  
  // Find your declination here: http://www.magnetic-declination.com/
  //#define COMPASS_DECLINATION_DEGREES 3.13 // 3°8' = 3.13 for Munich


  // Actual data (access them as you want, but they may get updated inside an interrupt)
  volatile float gpsCompass_magnet_x = 0.0f;
  volatile float gpsCompass_magnet_y = 0.0f;
  volatile float gpsCompass_magnet_z = 0.0f;
  //volatile float gpsCompass_heading_deg = 0.0f;


  // User callback
  typedef void (*t_gpsCompass_on_magnet)(float x, float y, float z); // We could also pass on all the data as parameters
  t_gpsCompass_on_magnet gpsCompass_on_magnet = NULL;
  /*
  typedef void (*t_gpsCompass_on_heading)(); // We could also pass on all the data as parameters
  t_gpsCompass_on_heading gpsCompass_on_heading = NULL;
  */

  // Used to signal between interrupt and main
  volatile bool gpsCompass_magnet_updated = false;
  //volatile bool gpsCompass_heading_updated = false;

  void gpsCompass_handle_compass() {
    /*
    // Called whenever there is compass data
    // Get a new sensor event
    sensors_event_t event;
    gpsMag.getEvent(&event);  // This hangs. It will block everything.
    */
    //// Manual I2C
    int x,y,z; //triple axis data
  
    // Tell the HMC5883L where to begin reading data
    Wire.beginTransmission(HMC5883_ADDRESS);
    Wire.write(0x03); //select register 3, X MSB register
    Wire.endTransmission();
    
   
    //Read data from each axis, 2 registers per axis
    Wire.requestFrom(HMC5883_ADDRESS, 6);
    if (Wire.available() >= 6){
      x = Wire.read()<<8; //X msb
      x |= Wire.read(); //X lsb
      
      z = Wire.read()<<8; //Z msb
      z |= Wire.read(); //Z lsb
      
      y = Wire.read()<<8; //Y msb
      y |= Wire.read(); //Y lsb

      float s = 0.001;
      gpsCompass_magnet_x = (float)x * s;
      gpsCompass_magnet_y = (float)y * s;
      gpsCompass_magnet_z = (float)z * s;
      gpsCompass_magnet_updated = true;
    }

    
    /*
    // Display the results (magnetic vector values are in micro-Tesla (uT))
    // Hold the module so that Z is pointing 'up' and you can measure the heading with x&y
    // Calculate heading when the magnetometer is level, then correct for signs of axis.
    float heading = atan2(event.magnetic.y, event.magnetic.x);

    // Declination
    heading += COMPASS_DECLINATION_DEGREES * DEG_TO_RAD;

    // Clip
    if (heading < 0)
      heading += 2*PI;
    else if (heading > 2*PI)
      heading -= 2*PI;

    // Convert radians to degrees for readability.
    heading = heading * RAD_TO_DEG;
    //gps_put_("Heading (deg): "); gps_put(heading);
    
    if (heading != gpsCompass_heading_deg) {
      gpsCompass_heading_deg = heading;
      gpsCompass_heading_updated = true;
    }
    */
    /*
    if (event.magnetic.x != gpsCompass_magnet_x) {
      gpsCompass_magnet_x = event.magnetic.x;
      gpsCompass_magnet_updated = true;
    }
    if (event.magnetic.y != gpsCompass_magnet_y) {
      gpsCompass_magnet_y = event.magnetic.y;
      gpsCompass_magnet_updated = true;
    }
    if (event.magnetic.z != gpsCompass_magnet_z) {
      gpsCompass_magnet_z = event.magnetic.z;
      gpsCompass_magnet_updated = true;
    }
    */
  }
#endif


void gpsCompass_loop() {
  /*
  while (gpsSerial.available() > 0) {
    if (gps.encode(gpsSerial.read())) {
      gpsCompass_handle_gps();
    }
  }
  */

  #ifdef USE_GPS
    //int timeout = TINYGPS_BUFFER_SIZE;
    
    // Because GPS interrupt returns quickly we have to trigger it from here where timing is non-critical
    while (tinyGpsBufOut != tinyGpsBufIn) {
      char c = tinyGpsBuf[tinyGpsBufOut];
      //gps_put_(c);
      
      if (gps.encode((char)c)) {
        gpsCompass_handle_gps();
      }
      tinyGpsBufOut = (tinyGpsBufOut + 1) % TINYGPS_BUFFER_SIZE;
      //timeout--;
    }

    if (gpsCompass_position_updated) {
      // Trigger user callback
      if (gpsCompass_on_position != NULL)
        gpsCompass_on_position(gpsCompass_position_lat, gpsCompass_position_lon, gpsCompass_position_alt);
      
      gpsCompass_position_updated = false;
    }
  #endif
  
  #ifdef USE_COMPASS
    // Compass outputs stuff all the time, so just trigger it
    //if (gpsCompass_magnet_found) {
      
      gpsCompass_handle_compass();
      
      if (gpsCompass_magnet_updated) {
        // Trigger user callback
        //if (gpsCompass_on_heading != NULL) gpsCompass_on_heading();
        if (gpsCompass_on_magnet != NULL) gpsCompass_on_magnet(gpsCompass_magnet_x, gpsCompass_magnet_y, gpsCompass_magnet_z);
        
        gpsCompass_magnet_updated = false;
      }
      /*
      if (gpsCompass_heading_updated) {
        // Trigger user callback
        if (gpsCompass_on_heading != NULL) gpsCompass_on_heading();
        gpsCompass_heading_updated = false;
      }
      */
    //}
  #endif
}



// Set up the serial port
void gpsCompass_setup(bool wireInit) {
  
  #ifdef USE_GPS
    // Start GPS
    
    /*
    // Clean reset
    pinMode(GPS_SERIAL_RX_PIN, OUTPUT);
    digitalWrite(GPS_SERIAL_RX_PIN, LOW);
    
    pinMode(GPS_SERIAL_TX_PIN, OUTPUT);
    digitalWrite(GPS_SERIAL_TX_PIN, LOW);
    
    delay(10);
    */

    gpsSerial.attachInterrupt(on_gpsSerial_rx);
    gpsSerial.begin(GPS_SERIAL_BAUD);

    /*
    // Initial handling
    for(int i = 0; i < 100; i++) {
      gpsCompass_loop();
      delay(1);
    }
    */
    
  //gps_put_(F("TinyGPS++ library v")); gps_put(TinyGPSPlus::libraryVersion());
  #endif


  #ifdef USE_COMPASS
    if (wireInit) Wire.begin();
  
    /*
    //// Adafruit
    if(gpsMag.begin()) {
      // Okay! Found
      gpsCompass_magnet_found = true;
      
    } else {
      // Magnetic sensor NOT found!
      gpsCompass_magnet_found = false;
      
    }
    */
    //// Manual I2C
    // Put the HMC5883 IC into the correct operating mode
    Wire.beginTransmission(HMC5883_ADDRESS); //open communication with HMC5883
    Wire.write(0x02); //select mode register
    Wire.write(0x00); //continuous measurement mode
    Wire.endTransmission();
  #endif
}


#endif
