#ifndef JOYSTICK_WCC
#define JOYSTICK_WCC
// Wii Classic Controller via I2C
#include <Wire.h>


#define WCC_I2C_ADDRESS 0x52

#define WCC_BYTE_COUNT (6)
#define WCC_I2C_ADDRESS 0x52

/*    Bit
  Byte  7 6 5 4 3 2 1 0
  0     RX<4:3>  |                  LX<5:0>
  1 RX<2:1>      |                  LY<5:0>
  2 RX<0> |   LT<4:3>   |           RY<4:0>
  3 LT<2:0>             |           RT<4:0>
  4 BDR BDD BLT B-  BH  B+  BRT Null
  5 BZL BB  BY  BA  BX  BZR BDL BDU
*/

struct WCCState {
  int8_t LeftX;
  int8_t LeftY;
  int8_t RightX;
  int8_t RightY;
  int8_t LeftT;
  int8_t RightT;
  boolean ButtonDown;
  boolean ButtonLeft;
  boolean ButtonUp;
  boolean ButtonRight;
  boolean ButtonSelect;
  boolean ButtonHome;
  boolean ButtonStart;
  boolean ButtonY;
  boolean ButtonX;
  boolean ButtonA;
  boolean ButtonB;
  boolean ButtonL;
  boolean ButtonR;
  boolean ButtonZL;
  boolean ButtonZR;
};

/**
   @brief Write bytes out to a device over I2C

   @param buf is a pointer to a string of bytes
   @param len is the length of the string
*/
void wcc_wire_write(uint8_t *buf, size_t len) {
  if ((buf != NULL) && (len > 0)) {
    Wire.beginTransmission(*buf++);
    while (--len != 0) {
      Wire.write(*buf++);
    }
    Wire.endTransmission();
  }
}

uint8_t wcc_init_string_1[] = {0x52, 0xf0, 0x55};
uint8_t wcc_init_string_2[] = {0x52, 0xfb, 0x00};
uint8_t wcc_zero_string[] = {0x52, 0x00};

void wcc_send_zero() {
  wcc_wire_write(wcc_zero_string, sizeof(wcc_zero_string));
}

void wcc_send_init() {
  wcc_wire_write(wcc_init_string_1, sizeof(wcc_init_string_1));
  wcc_wire_write(wcc_init_string_2, sizeof(wcc_init_string_2));
}

void wcc_decode_bytes(uint8_t *buf, size_t len, struct WCCState *myStruct) {

  if ((buf == NULL) || (len < WCC_BYTE_COUNT) || (myStruct == NULL)) {
    return;
  }

  myStruct->LeftX = (buf[0] & (64 - 1)) - 32; // 0 to 63
  myStruct->LeftY = (buf[1] & (64 - 1)) - 32; // 0 to 63 -> -32 to 31

  myStruct->RightX = (((buf[2] >> 7) & 1) + ((buf[1] >> 6) & 3) * 2 + ((buf[0] >> 6) & 3) * 8) - 16; // 0 to 31 -> -16 to 15
  myStruct->RightY = (buf[2] & (32 - 1)) - 16; // 0 to 31 -> -16 to 15

  //  Serial.print(myStruct->RightX);
  //  Serial.print('\n');
  //  Serial.print(myStruct->RightY);

  myStruct->LeftT = (((buf[2] >> 5) & 3) * 8 + ((buf[3] >> 5) & 7));
  myStruct->RightT = (buf[3] & (32 - 1));

  myStruct->ButtonDown = ((buf[4] & (1 << 6)) == 0);
  myStruct->ButtonLeft = ((buf[5] & (1 << 1)) == 0);
  myStruct->ButtonUp = ((buf[5] & (1 << 0)) == 0);
  myStruct->ButtonRight = ((buf[4] & (1 << 7)) == 0);
  myStruct->ButtonSelect = ((buf[4] & (1 << 4)) == 0);
  myStruct->ButtonHome = ((buf[4] & (1 << 3)) == 0);
  myStruct->ButtonStart = ((buf[4] & (1 << 2)) == 0);
  myStruct->ButtonY = ((buf[5] & (1 << 5)) == 0);
  myStruct->ButtonX = ((buf[5] & (1 << 3)) == 0);
  myStruct->ButtonA = ((buf[5] & (1 << 4)) == 0);
  myStruct->ButtonB = ((buf[5] & (1 << 6)) == 0);
  myStruct->ButtonL = ((buf[4] & (1 << 5)) == 0);
  myStruct->ButtonR = ((buf[4] & (1 << 1)) == 0);
  myStruct->ButtonZL = ((buf[5] & (1 << 7)) == 0);
  myStruct->ButtonZR = ((buf[5] & (1 << 2)) == 0);
}


void joystick_wcc_setup(bool wireInit) {
  if (wireInit)
    Wire.begin();
  
  wcc_send_init();
} 

struct WCCState wcc_state;

void joystick_wcc_loop() {
  uint8_t rawbytes[WCC_BYTE_COUNT];    // array to store arduino output
  size_t cnt = 0;

  wcc_send_zero(); // send the request for next bytes
  delay(20);
  
  Wire.requestFrom(WCC_I2C_ADDRESS, WCC_BYTE_COUNT);  // request data from nunchuck
  while (Wire.available()) {
    rawbytes[cnt++] = Wire.read();
  }

  // If we recieved the 6 bytes, then do something with them
  if (cnt >= WCC_BYTE_COUNT) {
    wcc_decode_bytes(rawbytes, cnt, &wcc_state);
  }
}
#endif
