#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
This is the Python counterpart for pybridge Arduino sketch.
It allows to control Arduino functions via serial. So it is something like "Firmata".

"""
import serial
import time

import thread
import Queue

SERIAL_BAUD = 115200
#SERIAL_BAUD = 230400
#SERIAL_BAUD = 250000
#SERIAL_BAUD = 256000

SHOW_TRAFFIC = not True
QUICK_RECOVERY = True


# Those constants must be kept in-sync with the pybridge Arduino firmware!
COMMAND_MAX = 0xff
#COMMAND_MAX_FLOOD = 4	# How many commands to send out at once at max. Keep in mind: Arduinos input buffer is small (64 bytes per default)
COMMAND_MAX_INPUT = 128	# How many bytes can the receiver buffer (64 bytes per default)

# CMD_CHAR  SEQ LEN  DATA... CHECK
# Offsets within one packet:
OFS_START = 0
OFS_SEQ = 1
OFS_LEN = 2
OFS_DATA = 3
OFS_CHECK = 4


CMD_CHAR = ord('!')
CMD_ACK = ord('A')
CMD_NACK = ord('N')
CMD_READY = ord('R')
CMD_DIR = ord('r')
CMD_SET_DIGITAL = ord('d')
CMD_SET_ANALOG = ord('a')
CMD_LEDRGB = ord('l')
CMD_GET = ord('g')
CMD_WATCH = ord('w')

KIND_DIGITAL = ord('D')
KIND_ANALOG = ord('A')
KIND_SERVO = ord('S')

DIR_INPUT = ord('I')
DIR_OUTPUT = ord('O')

DIGITALS = 12
ANALOGS = 4

VALUE_HIGH = 1
VALUE_LOW = 0
WATCH_ON = 1
WATCH_OFF = 0

WAIT_TIME = 0.005	# Time resolution/latency

def put(t):
	print('arduino:	' + str(t))

def dump(ba):
	r = ' (' + str(len(ba)) + ')	'
	for i in xrange(len(ba)):
		r += '%02X ' % int(ba[i])
	return r

class Pin:
	def __init__(self, arduino, kind, index, dir=None, valueDigital=None, valueAnalog=None, thresh=0, callback=None):	#, callbackData=None):
		self.arduino = arduino
		self.kind = kind
		self.index = index
		self.value = None
		#if (value): self.set(value)
		
		if (not valueDigital == None): self.setDigital(valueDigital)	# It's safer to set it before AND after
		if (not dir == None): self.dir(dir)
		if (not valueDigital == None): self.setDigital(valueDigital)
		
		if (not valueAnalog == None): self.setAnalog(valueAnalog)
		if (not thresh == None) and (callback): self.watch(thresh, callback)	#, callbackData)
		
	def dir(self, dir):
		self.arduino.dir(self.kind, self.index, dir)
	
	def setDigital(self, value):
		self.arduino.setDigital(self.kind, self.index, value)
	def setAnalog(self, value):
		self.arduino.setAnalog(self.kind, self.index, value)
	"""
	Obsolete: Use setAnalog
	def pwm(self, value):
		self.arduino.pwm(self.kind, self.index, value)
	
	Obsolete: Use setAnalog
	def servo(self, value):
		self.arduino.servo(self.kind, self.index, value)
	"""
	def get(self):
		"This returns the last known value, but asks to re-new it"
		self.arduino.get(self.kind, self.index)
		return self.value
		
	def watch(self, thresh=1, callback=None):	#, callbackData=None):
		"Start watching this value, call the callback once the value has moved more than the given threshold"
		def cb(value):	#(arduino, kind, index, value, cbData):
			self.value = value
			#callback(self, value, callbackData)
			callback(value)
		self.arduino.watch(self.kind, self.index, thresh, cb)	#, callbackData)
	
	def unwatch(self):
		self.arduino.watch(self.kind, self.index, thresh=0, cb=None)

"""
class LEDRGB:
	def __init__(self, arduino):
		self.arduino = arduino
	def setPixel(self, index, rgb):
		self.arduino.ledrgb(index, rgb)
"""

class Arduino:
	"Requires door2 sketch to be loaded onto the Arduino"
	
	#def Pin(self, kind, index, dir=None, value=None, thresh=1, callback=None):	#, callbackData=None):
	def Pin(self, kind, index, dir=None, valueDigital=None, valueAnalog=None, thresh=None, callback=None):	#, callbackData=None):
		#pin = Pin(arduino=self, kind=kind, index=index, dir=dir, value=value, thresh=thresh, callback=callback)	#, callbackData=callbackData)
		pin = Pin(arduino=self, kind=kind, index=index, dir=dir, valueDigital=valueDigital, valueAnalog=valueAnalog, thresh=thresh, callback=callback)	#, callbackData=callbackData)
		return pin
	
	def __init__(self, port):
		self.port = port
		self.baud = SERIAL_BAUD
		self.running = False
		self.ready = False
		self.callbacks = {
			KIND_DIGITAL: [None] * DIGITALS,
			KIND_ANALOG: [None] * ANALOGS
		}
		
		self.queue = Queue.Queue()
		self.seqLastSent = 0	# Last used sequence number (sent)
		self.seqLastAcked = 0	# Last sequence number ACKed by receiver
		self.dataQueue = {}	# Last data (that was not ack'ed)
		self.waitingBytes = 0	# Number of bytes waiting
		
		self.isBootLogging = False
		self.hasBootLog = False
		self.bootLogPackets = []
		put('Opening ' + port + '...')
		
		#if SIMULATION:
		#	put('*** RUNNING SIMULATION! HARDWARE NOT USED! ***')
		#	self.ser = None
		#else:
		#try:
		
		self.onError = None
		
		self.openConnection()
		#self.ser = serial.Serial(port=port, baudrate=115200, timeout=0, xonxoff=True, dsrdtr=True)
		
		#except serial.serialutil.SerialException as e:
		#	put('Error opening serial device ' + port + ': ' + str(e))
		#	return
		
		self.bootLog(False)	# Do not log this.
		
		self.startAndWait()
		
	def startAndWait(self):
		put('Watching Arduino on serial...')
		self.start_run()
		
		# Block execution until Arduino is ready...
		put('Waiting (blocking) until Arduino is ready...')
		while (not self.ready):
			time.sleep(0.1)
		
		put('Ready.')
		
	
	def reboot(self):
		put('reboot()...')
		
		self.stop()
		
		time.sleep(1)
		
		self.closeConnection()
		
		time.sleep(1)
		
		self.openConnection()
		
		self.startAndWait()
		
	
	def closeConnection(self):
		put('closeConnection()...')
		try:
			if not self.ser == None:
				self.ser.close()
		except Exception as e:
			put('Problem closing the serial port: ' + str(e))
		self.ser = None
	
	def openConnection(self):
		put('openConnection()...')
		#selfser.close()
		#self.ser = None
		self.ser = serial.Serial(port=self.port, baudrate=115200, timeout=0)
	
	def __del__(self):
		self.running = False
	
	def put(self, t):
		put('Arduino:\t' + t)
	
	"""
	#def pad(self, t):
	#	return str(t).zfill(COMMAND_DIGITS)
	def pad8(self, t):
		return '%0.2X' % t
	def pad16(self, t):
		return '%0.4X' % t
	def toBytearray16(value):
		return bytearray((value >> 8) % 256, value % 256)
	"""
	
	def start_run(self):
		self.running = True
		#self.run(self)
		thread.start_new_thread(self.run, (self,))
	
	def run(self, dummy):
		# Thread
		#commandString = ''
		rCmdState = OFS_START
		rCmdSeq = 0
		rCmdLen = 0
		rCmdCheckSupposed = 0
		
		rCmdCheckActual = 0
		rCmdData = []
		rCmdLenCurrent = 0
		
		dataNext = None
		lenNext = 0
		
		try:
			while self.running:
				wait = True
				
				#put('waitingBytes=%d' % self.waitingBytes)
				
				bytesToRead = self.ser.inWaiting()
				if (bytesToRead > 0):
					s = self.ser.read(bytesToRead)	# Read bytes
					if SHOW_TRAFFIC:
						put('<<<	Received: ' + dump(bytearray(s)) + '')
						put('<<<	          "' + str(s) + '"')
					
					for i in xrange(len(s)):
						c = ord(s[i])
						#put('#%d	%02X	state=%d	%s' % (i, c, rCmdState, chr(c)))
						
						# Protocol handling...
						
						if (rCmdState == OFS_START):
							if (c == CMD_CHAR):
								rCmdState = OFS_SEQ
							
						elif (rCmdState == OFS_SEQ):
							rCmdSeq = c
							rCmdState = OFS_LEN
						elif (rCmdState == OFS_LEN):
							rCmdLen = c
							rCmdCheckActual = 0xff
							rCmdData = []
							rCmdLenCurrent = 0
							if rCmdLen > 0: rCmdState = OFS_DATA
							else: rCmdState = OFS_CHECK
							
						elif (rCmdState == OFS_DATA):
							rCmdData.append(c)
							rCmdCheckActual = rCmdCheckActual ^ c
							rCmdLenCurrent += 1
							if rCmdLenCurrent >= rCmdLen:
								rCmdState = OFS_CHECK
						
						elif (rCmdState == OFS_CHECK):
							rCmdCheckSupposed = c
							if SHOW_TRAFFIC: put('handling...')
							self.handlePacket(rCmdSeq, rCmdLen, rCmdData, rCmdCheckSupposed, rCmdCheckActual)
							rCmdState = OFS_START
							
					wait = False
					#
				#
				
				# Thread safe queue
				#if (self.ready) and (not self.queue.empty()):
				if (self.ready):
					if (lenNext == 0) and (not self.queue.empty()):
							dataNext = self.queue.get()
							lenNext = len(dataNext)
							self.queue.task_done()
					
					if (lenNext > 0) and (self.waitingBytes + lenNext < COMMAND_MAX_INPUT):
						self.sendRaw(dataNext)
						dataNext = None
						lenNext = 0
						wait = False
				
				if wait: time.sleep(WAIT_TIME)
			#
		except Exception as e:
			self.error(e)
	
	def stop(self):
		self.running = False
	
	def wait(self):
		#while (not self.ready) or (self.waitingBytes > 0):
		while (not self.ready) or (self.waitingBytes > 0) or (self.queue.qsize() > 0) or (len(self.dataQueue) > 0):
			time.sleep(0.01)
		#
	#
	
	def error(self, e):
		put('error(): An error occured in Arduino lib: ' + str(e))
		if (not self.onError == None): self.onError(e)
		
		put('Rebooting...')
		self.reboot()
	
	def resync(self):
		"Re-start protocol"
		put('resync():	Stopping thread...')
		self.stop()
		
		put('resync():	Waiting...')
		time.sleep(0.25)
		
		put('resync():	Flushing queue...')
		while not self.queue.empty():
			try:
				self.queue.get(False)
			except Empty:
				continue
			self.queue.task_done()
		
		put('resync():	Waiting...')
		time.sleep(0.25)
		
		put('resync():	Flooding...')
		for i in xrange(100):
			self.ser.write(bytearray([0,0,0,0]))
			time.sleep(0.01)
		
		time.sleep(0.25)
		
		put('resync():	Flushing queue...')
		while not self.queue.empty():
			try:
				self.queue.get(False)
			except Empty:
				continue
			self.queue.task_done()
		
		put('resync():	Re-starting thread...')
		self.start_run()
		
		put('resync():	done.')
		
	
	def bootLog(self, active=False):
		if (active == False) and (len(self.bootLogPackets) > 0):
			self.hasBootLog = True
		self.isBootLogging = active
	
	def sendRaw(self, rawData):
		if SHOW_TRAFFIC:
			put('>>>	Sending ' + dump(rawData) + '...')
		
		if (OFS_SEQ < len(rawData)):
			nCmdSeq = int(rawData[OFS_SEQ])
			if nCmdSeq in self.dataQueue:
				# Packet was not acked since last overflow of sequence numbers. Just purge it...
				self.waitingBytes -= len(self.dataQueue[nCmdSeq])
		
		# Don't watch over the small ones... So: ACKs will be lost
		if (len(rawData) > 1+1+1+2+1):
			self.dataQueue[nCmdSeq] = rawData	# Store for re-sending until it is ack'ed
			self.waitingBytes += len(rawData)
		
		self.ser.write(rawData)
		#self.ser.flush()
		
	def sendPacket(self, cmdData):
		"Send raw payload, but take care of sequence, checksum, len"
		
		# Store commands for quick recovery after brown out or stuff
		if QUICK_RECOVERY:
			if self.isBootLogging:
				self.bootLogPackets.append(cmdData)
		
		# Construct packet
		cmdLen = len(cmdData)
		self.seqLastSent += 1
		cmdSeq = self.seqLastSent
		
		# Calc checksum
		cmdCheck = 0xff
		for i in xrange(cmdLen):
			cmdCheck ^= cmdData[i]
		
		data = bytearray([CMD_CHAR, cmdSeq % 0x100, cmdLen]) + cmdData + bytearray([cmdCheck])
		#data = bytearray([CMD_CHAR, cmdSeq % 0x100, cmdLen, cmdData, cmdCheck])
		
		# Actual raw data is ready!
		
		#if SHOW_TRAFFIC: put('Sending "' + data.strip() + '" (' + str(len(data)) + ')...')
		#self.ser.write(data)
		#self.dataLast = data
		
		if SHOW_TRAFFIC:	put('Queueing ' + dump(data.strip()) + '...')
		self.queue.put(data)
		
	
	def sendCommand_byte(self, cmd, value):
		self.sendPacket(bytearray([cmd, value % 0x100]))
	def sendCommand_pin_none(self, cmd, kind, index):
		self.sendPacket(bytearray([cmd, kind, index]))
	def sendCommand_pin_byte(self, cmd, kind, index, value):
		self.sendPacket(bytearray([cmd, kind, index, value % 0x100]))
	def sendCommand_pin_word(self, cmd, kind, index, value):
		self.sendPacket(bytearray([cmd, kind, index, (value >> 8) % 0x100, value % 0x100]))
	def sendCommand_pin_raw(self, cmd, kind, index, data):
		self.sendPacket(bytearray([cmd, kind, index]) + data)
	
	#def handleCommand(self, data):
	def handlePacket(self, cmdSeq, cmdLen, cmdData, cmdCheckSupposed, cmdCheckActual):
		"Handle the response packet coming back from Arduino"
		
		# Check integrity
		if not (cmdCheckSupposed == cmdCheckActual):
			self.sendCommand_byte(CMD_NACK, cmdSeq)
			return
		
		if (cmdLen == 0):
			if SHOW_TRAFFIC: put('handlePacket: No cmdLen! NACK!')
			return
		
		# Handle low level commands
		cmd = int(cmdData[0])
		
		if (cmd == CMD_ACK):
			nCmdSeq = int(cmdData[1])
			if SHOW_TRAFFIC: put('handlePacket: ACK #%02X' % nCmdSeq)
			self.seqLastAcked = nCmdSeq
			# Packet was successfully received
			# Clear from dataQueue
			if nCmdSeq in self.dataQueue:
				data = self.dataQueue[nCmdSeq]
				self.waitingBytes -= len(data)
				del self.dataQueue[nCmdSeq]
			return
			
		elif (cmd == CMD_NACK):
			nCmdSeq = int(cmdData[1])
			if SHOW_TRAFFIC: put('handlePacket: NACK #%02X! Re-sending...' % nCmdSeq)
			# Packet was not received correctly
			# Re-send!
			if not nCmdSeq in self.dataQueue:
				put('*** Packet was NACKed, but sequence number %d is not in queue %s ***' % (nCmdSeq, str(self.dataQueue)))
				#@TODO: Recover. And: decrement waitingBytes
			else:
				data = self.dataQueue[nCmdSeq]
				self.waitingBytes -= len(data)
				self.queue.put(data)
				#self.sendRaw(data)
			return
		
		# Ack the packet
		self.sendCommand_byte(CMD_ACK, cmdSeq)
		
		# Handle longer commands
		if (cmd == CMD_READY):
			put('handlePacket: pybridge reports ready!')
			self.ready = True
			
			# Quick recovery
			if QUICK_RECOVERY:
				if self.hasBootLog:
					put('handlePacket: Quick recovery (' + str(len(self.bootLogPackets)) + ' packets)...')
					for p in self.bootLogPackets:
						self.sendPacket(p)
					put('handlePacket: Quick recovery: done.')
			return
			
		else:
			
			# Handle pin commands
			if (cmdLen < 4):
				if SHOW_TRAFFIC: put('handlePacket: Too small! NACK!')
				self.sendCommand_byte(CMD_NACK, cmdSeq)
				return
			
			kind = int(cmdData[1])
			index = int(cmdData[2])
			if (cmdLen == 4):
				value = int(cmdData[3])
			elif (cmdLen == 5):
				value = int(cmdData[3]) << 8 + int(cmdData[4])
			else:
				value = cmdData[3:]
			
			self.handleCommand(cmd, kind, index, value)
		#
		
	
	def handleCommand(self, cmd, kind, index, value):
		if (cmd == CMD_GET):
			# Got a value...
			#put('Watch kind=' + kind + ', index=' + str(index) + ', value=' + str(value))
			
			#callbackTuple = self.callbacks[kind][index]
			#if (callbackTuple):
			#	callback = callbackTuple[0]
			#	if (callback): callback(self, kind, index, value, callbackTuple[1])
			callback = self.callbacks[kind][index]
			if (callback): callback(value)
			
		else:
			put('Unknown command %02X / "%s"!' % (cmd, chr(cmd)))
			#
	
	def dir(self, kind, index, dir):
		self.sendCommand_pin_byte(CMD_DIR, kind, index, dir)
	def setDigital(self, kind, index, value):
		self.sendCommand_pin_byte(CMD_SET_DIGITAL, kind, index, value)
	def setAnalog(self, kind, index, value):
		self.sendCommand_pin_word(CMD_SET_ANALOG, kind, index, value)
	
	def ledrgb_clear(self):
		self.sendCommand_pin_byte(CMD_LEDRGB, 'C', 0,0)
	def ledrgb_pixel(self, index, rgb=[0,0,0]):
		self.sendCommand_pin_raw(CMD_LEDRGB, 'P', index, data=bytearray([rgb[0], rgb[1], rgb[2]]))
	def ledrgb_raw(self, pixels):
		"""
		data = ''
		asciiStart = 32
		asciiRange = 64
		asciiScale = 256 / asciiRange
		for i in xrange(len(pixels)):
			rgb = pixels[i]
			data += chr(asciiStart + int(rgb[0]/asciiScale))
			data += chr(asciiStart + int(rgb[1]/asciiScale))
			data += chr(asciiStart + int(rgb[2]/asciiScale))
			
		self.cmd(CMD_LEDRGB, 'R', len(pixels), data)
		"""
		self.sendCommand_pin_raw(CMD_LEDRGB, 'R', len(pixels), bytearray(pixels))
		
	def ledrgb_fill(self, count, rgb=[0,0,0]):
		self.sendCommand_pin_raw(CMD_LEDRGB, 'F', count, bytearray(rgb))
	def ledrgb_update(self):
		self.sendCommand_pin_none(CMD_LEDRGB, 'S', COMMAND_MAX)
		
	def ledrgb_blackOut(self):
		"Just a little helper to clear and commit"
		self.ledrgb_clear()
		self.ledrgb_update()
	
	def get(self, kind, index):
		self.sendCommand_pin_none(CMD_GET, kind, index)
	def watch(self, kind, index, thresh=WATCH_ON, callback=None):	#, callbackData=None):
		#self.callbacks[kind][index] = (callback, callbackData)
		self.callbacks[kind][index] = callback
		self.sendCommand_pin_word(CMD_WATCH, kind, index, thresh)
	

if __name__ == '__main__':
	# Test it
	put('Running test...')
	SHOW_TRAFFIC = True
	
	items = [
		CMD_CHAR,
		CMD_ACK,
		CMD_NACK,
		CMD_READY,
		CMD_DIR,
		CMD_SET_DIGITAL,
		CMD_SET_ANALOG,
		CMD_LEDRGB,
		CMD_GET,
		CMD_WATCH,
		
		KIND_DIGITAL,
		KIND_ANALOG,
		KIND_SERVO,
		#KIND_LEDRGB,
	]
	for i in items:
		put('%s	%02X' % (chr(i), i))
	#sys.exit()
	
	
	#arduino = Arduino('COM57')
	arduino = Arduino('COM3')
	
	"""
	# Protocol flood test
	#arduino.cmd(cmd, kind, index, value=None)
	time.sleep(1)
	
	put('Flooding...')
	for i in xrange(300):
		#put(str(arduino.dataQueue))
		#arduino.sendCommand_pin_none(cmd=CMD_DIR, kind='T', index=1)
		arduino.sendCommand_pin_raw(cmd=CMD_DIR, kind='T', index=1, data=bytearray([1,2,3,4,5,6,7,8,9,10,11,12]))
		#time.sleep(0.01)
	
	put('Waiting for all packets to be ACKed...')
	put(str(arduino.dataQueue))
	while arduino.waitingBytes > 0:
		put('Still waiting for %d + %d packets to be acked (%d bytes currently underway)...' % (arduino.queue.qsize(), len(arduino.dataQueue), arduino.waitingBytes))
		time.sleep(0.01)
	arduino.wait()
	put('Flooding ended cleanly.')
	
	#time.sleep(1)
	#arduino.cmd(cmd=CMD_DIR, kind='T', index=1, value=64)
	"""
	
	"""
	# Manual mode
	arduino.dir(KIND_DIGITAL, 0, DIR_INPUT)
	arduino.set(KIND_DIGITAL, 0, VALUE_HIGH)
	arduino.watch(KIND_DIGITAL, 0, WATCH_ON)
	"""
	
	# Super-short
	def myCallback(value):
		put('myCallback! The pin has changed to ' + str(value) + '!')
		
	pin = arduino.Pin(kind=KIND_DIGITAL, index=0, dir=DIR_INPUT, valueDigital=VALUE_HIGH, thresh=WATCH_ON, callback=myCallback)
	#pin = arduino.Pin(kind=KIND_DIGITAL, index=1, dir=DIR_INPUT, valueDigital=VALUE_HIGH, thresh=WATCH_ON, callback=myCallback)
	#pin = arduino.Pin(kind=KIND_DIGITAL, index=2, dir=DIR_INPUT, valueDigital=VALUE_HIGH, thresh=WATCH_ON, callback=myCallback)
	
	time.sleep(10)
	put('Test ended (timeout)')

