#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
This is the Python counterpart for pybridge Arduino sketch.
It allows to control Arduino functions via serial. So it is something like "Firmata".


 * Protocol:
 *   * Commands are sent using a simple package structure that takes care of checksums, flow control etc.
 *   * Packet layout:
 *      MAGIC  CMD  SEQ [LEN  PAYLOAD...] CHECK
 *      |      |    |    |    |           |
 *      |      |    |    |    |           |__ BYTE: Checksum. Simply "0xFF XOR <all bytes after MAGIC>"
 *      |      |    |    |    |
 *      |      |    |    |    |__ BYTEARRAY: Payload bytes of variable length.
 *      |      |    |    |
 *      |      |    |    |__ BYTE: Length of payload
 *      |      |    |
 *      |      |    |__ BYTE: Sequence number
 *      |      |
 *      |      |__ BYTE: Command
 *      |      
 *      |
 *      |__ BYTE: Start byte (PACKET_MAGIC) for synchronization, usually "!"
 * 

#@TODO: Re-send automatically if no ACK in X milliseconds

"""
import serial
import time

import thread
from Queue import Queue
from threading import Lock

SERIAL_BAUD = 115200	# Default baudrate
#SERIAL_BAUD = 230400
#SERIAL_BAUD = 250000
#SERIAL_BAUD = 256000

SHOW_TRAFFIC = not True
SHOW_TRAFFIC_TEXT = SHOW_TRAFFIC
QUICK_RECOVERY = True


# Those constants must be kept in-sync with the pybridge Arduino firmware!
COMMAND_MAX = 0xff
#COMMAND_MAX_INPUT = 128	# How many bytes can the receiver buffer (64 bytes per default)
COMMAND_MAX_INPUT = 32	# How many bytes can the receiver buffer (64 bytes per default)
COMMAND_TIMEOUT = 0.2	# Timeout inside a received packet

PACKET_MAGIC = ord('!')

CMD_ACK = ord('A')
CMD_NACK = ord('N')
CMD_UNKNOWN = ord('?')
CMD_ERROR = ord('X')
CMD_READY = ord('R')
VOLATILE_COMMANDS = [CMD_ACK, CMD_NACK, CMD_UNKNOWN, CMD_ERROR]	# They can not be re-sent

CMD_DIR = ord('r')
CMD_SET_DIGITAL = ord('d')
CMD_SET_ANALOG = ord('a')
CMD_GET = ord('g')
CMD_PUT = ord('p')
CMD_WATCH = ord('w')
CMD_LEDRGB = ord('l')

KIND_DIGITAL = ord('D')
KIND_ANALOG = ord('A')
KIND_SERVO = ord('S')

DIR_INPUT = ord('I')
DIR_OUTPUT = ord('O')

# Error codes
ERROR_LENGTH = 1	# Error about payload size
ERROR_SEQ_DELTA = 2	# Acked and last acked sequence numbers differ more than 1
ERROR_OBSOLETE = 3	# NACK requested a packet that is not available any more


# Protocol states
# PACKET_MAGIC  SEQ LEN  DATA... CHECK
# Offsets within one packet:
OFS_START = 0
OFS_CMD = 1
OFS_SEQ = 2
OFS_LEN = 3
OFS_DATA = 4
OFS_CHECK = 5


DIGITALS = 12
ANALOGS = 4

VALUE_HIGH = 1
VALUE_LOW = 0
WATCH_ON = 1
WATCH_OFF = 0

WAIT_TIME = 0.005	# Time resolution/latency

def put(t):
	print('arduino:	' + str(t))

def dump(ba):
	if (ba is None): return 'NONE'
	
	r = '(' + str(len(ba)) + ')	'
	for i in xrange(len(ba)):
		r += '%02X ' % int(ba[i])
	return r

class Pin:
	def __init__(self, arduino, kind, index, dir=None, valueDigital=None, valueAnalog=None, thresh=0, callback=None):	#, callbackData=None):
		self.arduino = arduino
		self.kind = kind
		self.index = index
		self.value = None
		#if (value): self.set(value)
		
		if (not valueDigital == None): self.setDigital(valueDigital)	# It's safer to set it before AND after
		if (not dir == None): self.dir(dir)
		if (not valueDigital == None): self.setDigital(valueDigital)
		
		if (not valueAnalog == None): self.setAnalog(valueAnalog)
		if (not thresh == None) and (callback): self.watch(thresh, callback)	#, callbackData)
		
	def dir(self, dir):
		self.arduino.dir(self.kind, self.index, dir)
	
	def setDigital(self, value):
		self.arduino.setDigital(self.kind, self.index, value)
	def setAnalog(self, value):
		self.arduino.setAnalog(self.kind, self.index, value)
	"""
	Obsolete: Use setAnalog
	def pwm(self, value):
		self.arduino.pwm(self.kind, self.index, value)
	
	Obsolete: Use setAnalog
	def servo(self, value):
		self.arduino.servo(self.kind, self.index, value)
	"""
	def get(self):
		"This returns the last known value, but asks to re-new it"
		self.arduino.get(self.kind, self.index)
		return self.value
		
	def watch(self, thresh=1, callback=None):	#, callbackData=None):
		"Start watching this value, call the callback once the value has moved more than the given threshold"
		def cb(value):	#(arduino, kind, index, value, cbData):
			self.value = value
			#callback(self, value, callbackData)
			callback(value)
		self.arduino.watch(self.kind, self.index, thresh, cb)	#, callbackData)
	
	def unwatch(self):
		self.arduino.watch(self.kind, self.index, thresh=0, cb=None)

"""
class LEDRGB:
	def __init__(self, arduino):
		self.arduino = arduino
	def setPixel(self, index, rgb):
		self.arduino.ledrgb(index, rgb)
"""

class Arduino:
	"Requires pybridge sketch to be loaded onto the Arduino"
	
	
	def __init__(self, port, baud=SERIAL_BAUD, onCommand=None):
		self.port = port
		self.baud = baud
		self.running = False
		self.ready = False
		self.callbacks = {
			KIND_DIGITAL: [None] * DIGITALS,
			KIND_ANALOG: [None] * ANALOGS
		}
		
		self.queue = Queue()
		self.packetLock = Lock()
		self.packetSeqLastSent = 0	# Last used sequence number (sent)
		self.packetSeqLastAcked = 0	# Last sequence number ACKed by receiver
		self.dataQueue = {}	# Last data (that was not ack'ed)
		self.waitingBytes = 0	# Number of bytes waiting
		
		
		self.isBootLogging = False
		self.hasBootLog = False
		self.bootLogPackets = []
		
		
		#put('Opening ' + port + '...')
		
		#if SIMULATION:
		#	put('*** RUNNING SIMULATION! HARDWARE NOT USED! ***')
		#	self.ser = None
		#else:
		#try:
		
		self.onError = None
		self.onPacket = None
		self.onCommand = onCommand
		
		self.openConnection()
		#self.ser = serial.Serial(port=port, baudrate=115200, timeout=0, xonxoff=True, dsrdtr=True)
		
		#except serial.serialutil.SerialException as e:
		#	put('Error opening serial device ' + port + ': ' + str(e))
		#	return
		
		self.bootLog(False)	# Do not log this.
		
		self.startAndWait()
		
	def startAndWait(self):
		put('startAndWait()...')
		self.start_run()
		
		# Block execution until Arduino is ready...
		put('startAndWait: Waiting (blocking) until Arduino is ready...')
		while (not self.ready):
			time.sleep(0.1)
		
		put('startAndWait: Ready.')
		
	
	def reboot(self):
		put('reboot()...')
		
		self.stop()
		
		time.sleep(1)
		
		self.closeConnection()
		
		time.sleep(1)
		
		self.openConnection()
		
		self.startAndWait()
		
	
	def closeConnection(self):
		put('closeConnection()...')
		try:
			if not self.ser == None:
				self.ser.close()
		except Exception as e:
			put('closeConnection: Problem closing the serial port: ' + str(e))
		self.ser = None
	
	def openConnection(self):
		put('openConnection()...')
		#selfser.close()
		#self.ser = None
		self.ser = serial.Serial(port=self.port, baudrate=self.baud, timeout=0)
	
	def __del__(self):
		self.running = False
	
	def put(self, t):
		put('Arduino:\t' + t)
	
	"""
	#def pad(self, t):
	#	return str(t).zfill(COMMAND_DIGITS)
	def pad8(self, t):
		return '%0.2X' % t
	def pad16(self, t):
		return '%0.4X' % t
	def toBytearray16(value):
		return bytearray((value >> 8) % 256, value % 256)
	"""
	
	def start_run(self):
		self.running = True
		#self.run(self)
		thread.start_new_thread(self.run, (self,))
	
	def run(self, dummy):
		# Thread
		#commandString = ''
		rPacketState = OFS_START
		rPacketCmd = 0
		rPacketSeq = 0
		rPacketPayloadLen = 0
		rPacketPayloadLenCurrent = 0
		rPacketPayload = []
		
		rPacketCheckSupposed = 0
		rPacketCheckActual = 0
		
		rPacketStartTime = time.time()
		
		dataNext = None
		lenNext = 0
		
		#try:
		if True:
			while self.running:
				wait = True	# If no one explicitly requests not to, we wait a little while at the end of that loop to keep CPU usage low
				
				#put('waitingBytes=%d' % self.waitingBytes)
				
				bytesToRead = self.ser.inWaiting()
				if (bytesToRead == 0):
					# No new data
					
					if (rPacketState > OFS_START):
						# We are currently inside a command
						
						# Check if we have a timeout
						now = time.time()
						delay = now - rPacketStartTime
						if (delay > COMMAND_TIMEOUT):
							# Timeout has occured
							put('Timeout - Resetting protocol state')
							
							# Reset state
							rPacketState = OFS_START
							
							# If we already have a sequence number: NACK it!
							if (rPacketState > OFS_SEQ):
								if SHOW_TRAFFIC: put('NACKing')
								self.sendCommand_byte(CMD_NACK, rPacketSeq);
							
						
					
				else:
					# There are bytes waiting
					
					# Actually read from serial
					s = self.ser.read(bytesToRead)
					
					if SHOW_TRAFFIC:
						#put('<<<	Received (%d): %s' % (len(s), dump(bytearray(s))))
						put('<<<	Received	%s' % (dump(bytearray(s))))
						if SHOW_TRAFFIC_TEXT:
							put('<<<	        	   	"' + str(s.strip()) + '"')
					
					for i in xrange(len(s)):
						c = ord(s[i])
						#put('#%d	%02X	state=%d	%s' % (i, c, rPacketState, chr(c)))
						
						# Checksum
						#put('%02X xor %02X = %02x' % (rPacketCheckActual, c, rPacketCheckActual^c ))
						rPacketCheckActual ^= c
						
						if (rPacketState == OFS_START):
							if (c == PACKET_MAGIC):
								# Record start of packet
								rPacketStartTime = time.time()
								
								# Reset packet memory
								rPacketSeq = 0
								rPacketPayloadLen = 0
								#rPacketPayloadLenCurrent = 0
								rPacketPayload = None
								rPacketCheckActual = 0xff	# Reset checksum
								rPacketCheckSupposed = 0
								
								rPacketState = OFS_CMD
							else:
								# Not magic: Ignore / discard byte
								if SHOW_TRAFFIC:
									put('*** Out of bounds byte (not magic): 0x%02X ***' % (c))
								pass
							
						elif (rPacketState == OFS_CMD):
							rPacketCmd = c
							rPacketState = OFS_SEQ
						
						elif (rPacketState == OFS_SEQ):
							rPacketSeq = c
							
							# Short volatile packets without payload end here
							if rPacketCmd in VOLATILE_COMMANDS:
								# No payload needed
								rPacketState = OFS_CHECK	# Straight to checksum
							else:
								# Continue with payload
								rPacketState = OFS_LEN
							
						elif (rPacketState == OFS_LEN):
							rPacketPayloadLen = c
							rPacketPayload = []
							rPacketPayloadLenCurrent = 0
							if rPacketPayloadLen > 0: rPacketState = OFS_DATA
							else: rPacketState = OFS_CHECK
							
						elif (rPacketState == OFS_DATA):
							rPacketPayload.append(c)
							rPacketPayloadLenCurrent += 1
							if rPacketPayloadLenCurrent >= rPacketPayloadLen:
								rPacketState = OFS_CHECK
						
						elif (rPacketState == OFS_CHECK):
							rPacketCheckSupposed = c
							# Since the checksumActual is also XORed with this last byte, it should be zero if correct!
							self.handlePacket(rPacketCmd, rPacketSeq, rPacketPayloadLen, rPacketPayload, rPacketCheckActual)
							rPacketState = OFS_START
							
					wait = False	# If we had bytes now, we likely have more. So do not wait.
					#
				#
				
				
				# Thread safe sending
				if (self.ready):
					
					# Check if there are more things to be sent on queue
					if (lenNext == 0) and (not self.queue.empty()):
						# Get more data from queue
						dataNext = self.queue.get()
						lenNext = len(dataNext)
						self.queue.task_done()
					
					# Do we have data to be sent? Does it fit the output buffer?
					if (lenNext > 0):
						if ((self.waitingBytes + lenNext) < COMMAND_MAX_INPUT):
							# We have data to be sent and the serial buffer could take it
							self.sendRaw(dataNext)
							self.waitingBytes += lenNext	# Not for volatile ones!
							dataNext = None
							lenNext = 0
							wait = False	# Sending might have took a while. So no need for more waiting
						else:
							put('*** COMMAND_MAX_INPUT (%d of %d) reached ***' % (self.waitingBytes, COMMAND_MAX_INPUT))
				
				if wait: time.sleep(WAIT_TIME)
			#
		#except Exception as e:
		#	self.error(e)
	
	def stop(self):
		self.running = False
	
	def wait(self):
		#while (not self.ready) or (self.waitingBytes > 0):
		while (not self.ready) or (self.waitingBytes > 0) or (self.queue.qsize() > 0) or (len(self.dataQueue) > 0):
			time.sleep(0.01)
		#
	#
	
	def error(self, e):
		put('error(): An error occured in Arduino lib: ' + str(e))
		if (not self.onError == None): self.onError(e)
		
		put('Rebooting arduino...')
		self.reboot()
	
	def resync(self):
		"Re-start protocol"
		put('resync():	Stopping thread...')
		self.stop()
		
		put('resync():	Waiting...')
		time.sleep(COMMAND_TIMEOUT)
		
		put('resync():	Flushing queue...')
		while not self.queue.empty():
			try:
				self.queue.get(False)
				self.queue.task_done()
			except Exception as e:
				continue
		
		put('resync():	Waiting...')
		time.sleep(COMMAND_TIMEOUT*2)
		
		"""
		put('resync():	Flooding...')
		for i in xrange(100):
			self.ser.write(bytearray([0,0,0,0]))
			time.sleep(0.01)
		
		time.sleep(0.25)
		
		put('resync():	Flushing queue...')
		while not self.queue.empty():
			try:
				self.queue.get(False)
			except Empty:
				continue
			self.queue.task_done()
		"""
		
		put('resync():	Re-starting thread...')
		self.start_run()
		
		put('resync():	done.')
		
	
	def bootLog(self, active=False):
		if (active == False) and (len(self.bootLogPackets) > 0):
			self.hasBootLog = True
		self.isBootLogging = active
	
	def queueRaw(self, data):
		if SHOW_TRAFFIC:
			put('>>>	Queueing	%s' % (dump(data)))
		self.queue.put(data)
	
	def sendRaw(self, rawData):
		if SHOW_TRAFFIC:
			#put('>>>	Sending (%d): %s' % (len(rawData), dump(rawData)))
			put('>>>	Sending 	%s' % (dump(rawData)))
			if SHOW_TRAFFIC_TEXT:
				put('>>>	        	   	"%s"' % (str(rawData).strip()))
		
		self.ser.write(rawData)
		#self.ser.flush()
	
	def sendPacket_volatile(self, cmd, seq):
		"Send a short packet, like ACK (uses sequence number as parameter, no re-sending possible)"
		check = 0xff ^ cmd ^ seq
		data = bytearray([PACKET_MAGIC, cmd, seq, check])
		
		self.queueRaw(data)
		
		# Volatile packets are not ACKed. So we need to remove their size from waitingBytes
		self.waitingBytes -= len(data)	# We do not wait for those
		
	def sendPacket_safe(self, cmd, payload=None):
		"Send standard packet (increments sequence number)"
		
		# Store setup while booting for a quick recovery after brown out or stuff
		
		if QUICK_RECOVERY:
			if self.isBootLogging:
				self.bootLogPackets.append((cmd, payload))
		
		# Construct packet
		if (payload is None): payload = []
		payloadLen = len(payload)
		
		self.packetLock.acquire()
		
		# Increment sequence counter
		seq = (self.packetSeqLastSent + 1) % 0x100
		
		# Calc checksum
		check = 0xff ^ cmd ^ seq ^ payloadLen
		for i in xrange(payloadLen):
			check ^= payload[i]
		
		data = bytearray([PACKET_MAGIC, cmd, seq, payloadLen]) + bytearray(payload) + bytearray([check])
		#data = bytearray([PACKET_MAGIC, packetSeq % 0x100, packetPayloadLen, packetPayload, packetCheck])
		
		# Actual raw data is ready!
		self.packetSeqLastSent = seq
		
		self.packetLock.release()
		
		# Store for potential re-sending
		self.dataQueue[seq] = data
		
		# Queue for transmission
		self.queueRaw(data)
		
	
	def sendCommand_none(self, cmd):
		"Send a command with no further data"
		self.sendPacket_safe(cmd)
	def sendCommand_data(self, cmd, data):
		"Send a command with no further data"
		self.sendPacket_safe(cmd, data)
	def sendCommand_byte(self, cmd, value):
		self.sendPacket_safe(cmd, [value % 0x100])
	def sendCommand_pin_none(self, cmd, kind, index):
		self.sendPacket_safe(cmd, [kind, index])
	def sendCommand_pin_byte(self, cmd, kind, index, value):
		self.sendPacket_safe(cmd, [kind, index, value % 0x100])
	def sendCommand_pin_word(self, cmd, kind, index, value):
		self.sendPacket_safe(cmd, [kind, index, (value >> 8) % 0x100, value % 0x100])
	def sendCommand_pin_raw(self, cmd, kind, index, data):
		self.sendPacket_safe(cmd, [kind, index] + data)
	
	def handlePacket(self, cmd, seq, payloadLen, payload, checkActual):
		"Handle the response packet coming back from Arduino"
		
		#if SHOW_TRAFFIC: put('handlePacket()...')
		
		# Check integrity (checksum)
		#if not (checkSupposed == checkActual):
		if not (checkActual == 0):
			#if SHOW_TRAFFIC:
			put('handlePacket: *** Checksums differ (%02X > 0 in %s)! NACKing #%02X! ***' % (checkActual, dump([PACKET_MAGIC, cmd, seq, payloadLen] + payload), seq))
			self.sendPacket_volatile(CMD_NACK, seq)
			return
		
		# Custom packet handler
		if (self.onPacket is not None):
			if self.onPacket(cmd, seq, payload):
				return	# Stop propagation if result is True
		
		# Handle low level commands
		if (cmd == CMD_ACK):
			# Packet was received. Good.
			
			if SHOW_TRAFFIC: put('handlePacket: Got ACK #%02X' % (seq))
			
			# Clear from dataQueue
			if seq in self.dataQueue:
				data = self.dataQueue[seq]
				self.waitingBytes -= len(data)
				del self.dataQueue[seq]
			else:
				put('handlePacket: *** ACK #%02X, but not in queue, ignoring ***' % (seq))
				return	# Don't ACK the ACK
			
			# Check if a number was skipped
			delta = seq - self.packetSeqLastAcked
			# Handle wrap-around (e.g. lastAcked=0xff, nCmdSeq=0x02)
			if (delta < -0x7f):
				delta += 0x100
			
			if (delta > 1):
				# Seems like ACK skipped a packet!
				put('handlePacket: *** ACK #%02X differs from last one #%02X too much (%d > 1) ***' % (seq, self.packetSeqLastAcked, delta))
				put('handlePacket: *** The intermediate packet(s) (#%02X - #%02X) might got lost! ***' % (self.packetSeqLastAcked+1, seq-1))
				
				# Re-send the skipped ones
				seqResend = (self.packetSeqLastAcked+1) % 0x100	# Where to start
				for i in xrange(delta-1):
					if seqResend in self.dataQueue:
						put('handlePacket: Re-sending skipped #%02X' % (seqResend))
						data = self.dataQueue[seqResend]
						self.waitingBytes -= len(data)	# It was lost, so the bytes are not waiting any more
						# Queue for re-transmission
						self.queueRaw(data)
					else:
						# This skipped packet is not in queue, it might have been ACKed already earlier
						if SHOW_TRAFFIC:
							put('handlePacket: Not re-sending skipped #%02X' % (seqResend))
						
					seqResend = (seqResend + 1) % 0x100
				
				# Continue as usual
				self.packetSeqLastAcked = seq
			elif (delta < 0):
				# ACK of an older packet (maybe a re-transmitted one)
				put('handlePacket: ACK of old packet #%02X (maybe a re-transmitted one)' % (seq))
				# Do NOT set the packetSeqLastAcked to that one!
				
			else:
				# Everything is in order. Continue with that one.
				self.packetSeqLastAcked = seq
			
			return	# Don't ACK the ACK
			
		elif (cmd == CMD_NACK):
			if SHOW_TRAFFIC:
				put('handlePacket: Got NACK #%02X!' % seq)
			
			# Packet was not received correctly
			if not seq in self.dataQueue:
				# This may happen for "too short" packets which do not get stored
				put('handlePacket: *** Packet #%02X was NACKed, but is not in queue ***' % (seq))	#, str(self.dataQueue)))
				
			else:
				# Re-send!
				put('handlePacket: Re-sending NACKed #%02X...' % seq)
				data = self.dataQueue[seq]
				self.waitingBytes -= len(data)	# NACKed data is also removed (because it was transmitted)
				self.queueRaw(data)
			return	# Don't ACK the NACK
			
		elif (cmd == CMD_UNKNOWN):
			put('handlePacket: *** CMD_UNKNOWN %02X / %s ***' % (seq, dump(payload)))
			return	# Don't ACK the UNKNOWN
			
		elif (cmd == CMD_ERROR):
			put('handlePacket: *** CMD_ERROR %02X / %s ***' % (seq, dump(payload)))
			return	# Don't ACK the UNKNOWN
			
		
		# Ack the packet
		if SHOW_TRAFFIC: put('handlePacket: ACKing #%02X...' % (seq))
		self.sendPacket_volatile(CMD_ACK, seq)
		
		
		# Handle longer commands
		if (cmd == CMD_READY):
			put('handlePacket: CMD_READY!')
			self.ready = True
			
			# Quick recovery
			if QUICK_RECOVERY:
				if self.hasBootLog:
					put('handlePacket: Quick recovery (' + str(len(self.bootLogPackets)) + ' packets)...')
					for p in self.bootLogPackets:
						self.sendPacket(p)
					put('handlePacket: Quick recovery done.')
			return
		
		
		# Handle custom commands
		if (self.onCommand is not None):
			if self.onCommand(cmd, payload):
				return	# Stop propagation if result is True
		
		
		# Handle built-in commands
		self.handleCommand(cmd, payload)
		
		
	
	def handleCommand(self, cmd, data):
		if SHOW_TRAFFIC: put('handleCommand()...')
		
		l = len(data)
		
		# Handle "pin" commands
		#if cmd in [CMD_DIR, CMD_SET_DIGITAL, CMD_SET_ANALOG, CMD_GET, CMD_WATCH, CMD_LEDRGB]
		if cmd in [CMD_PUT]:	# All "pin" commands we accept from arduino
			if (l >= 2):
				# Extract actual "pin parameters"
				
				kind = int(data[0])
				index = int(data[1])
				if (l == 3):
					value = int(data[2])
				elif (l == 4):
					value = int(data[2]) << 8 + int(data[3])
				else:
					value = data[2:]
				
				self.handleCommand_pin(cmd, kind, index, value)
			#
		else:
			put('handleComand: Received unknown command from Arduino: 0x%02X' % (cmd))
			pass
		
	#
	
	#def Pin(self, kind, index, dir=None, value=None, thresh=1, callback=None):	#, callbackData=None):
	def Pin(self, kind, index, dir=None, valueDigital=None, valueAnalog=None, thresh=None, callback=None):	#, callbackData=None):
		#pin = Pin(arduino=self, kind=kind, index=index, dir=dir, value=value, thresh=thresh, callback=callback)	#, callbackData=callbackData)
		pin = Pin(arduino=self, kind=kind, index=index, dir=dir, valueDigital=valueDigital, valueAnalog=valueAnalog, thresh=thresh, callback=callback)	#, callbackData=callbackData)
		return pin
	
	def handleCommand_pin(self, cmd, kind, index, value):
		if (cmd == CMD_PUT):
			# Got a value...
			#put('Watch kind=' + kind + ', index=' + str(index) + ', value=' + str(value))
			
			#callbackTuple = self.callbacks[kind][index]
			#if (callbackTuple):
			#	callback = callbackTuple[0]
			#	if (callback): callback(self, kind, index, value, callbackTuple[1])
			callback = self.callbacks[kind][index]
			if (callback): callback(value)
			
		else:
			put('Unknown command %02X / "%s"!' % (cmd, chr(cmd)))
			#
	
	
	# Custom commands
	def dir(self, kind, index, dir):
		self.sendCommand_pin_byte(CMD_DIR, kind, index, dir)
	def setDigital(self, kind, index, value):
		self.sendCommand_pin_byte(CMD_SET_DIGITAL, kind, index, value)
	def setAnalog(self, kind, index, value):
		self.sendCommand_pin_word(CMD_SET_ANALOG, kind, index, value)
	
	def get(self, kind, index):
		self.sendCommand_pin_none(CMD_GET, kind, index)
	def watch(self, kind, index, thresh=WATCH_ON, callback=None):	#, callbackData=None):
		#self.callbacks[kind][index] = (callback, callbackData)
		self.callbacks[kind][index] = callback
		self.sendCommand_pin_word(CMD_WATCH, kind, index, thresh)
	
	
	# LED RGB stuff
	def ledrgb_clear(self):
		self.sendCommand_pin_byte(CMD_LEDRGB, 'C', 0,0)
	def ledrgb_pixel(self, index, rgb=[0,0,0]):
		self.sendCommand_pin_raw(CMD_LEDRGB, 'P', index, data=bytearray([rgb[0], rgb[1], rgb[2]]))
	def ledrgb_raw(self, pixels):
		"""
		data = ''
		asciiStart = 32
		asciiRange = 64
		asciiScale = 256 / asciiRange
		for i in xrange(len(pixels)):
			rgb = pixels[i]
			data += chr(asciiStart + int(rgb[0]/asciiScale))
			data += chr(asciiStart + int(rgb[1]/asciiScale))
			data += chr(asciiStart + int(rgb[2]/asciiScale))
			
		self.cmd(CMD_LEDRGB, 'R', len(pixels), data)
		"""
		self.sendCommand_pin_raw(CMD_LEDRGB, 'R', len(pixels), bytearray(pixels))
		
	def ledrgb_fill(self, count, rgb=[0,0,0]):
		self.sendCommand_pin_raw(CMD_LEDRGB, 'F', count, bytearray(rgb))
	def ledrgb_update(self):
		self.sendCommand_pin_none(CMD_LEDRGB, 'S', COMMAND_MAX)
		
	def ledrgb_blackOut(self):
		"Just a little helper to clear and commit"
		self.ledrgb_clear()
		self.ledrgb_update()
	

if __name__ == '__main__':
	# Test it
	
	put('Running test (verbose traffic)...')
	SHOW_TRAFFIC = True
	SHOW_TRAFFIC_TEXT = True
	
	"""
	items = [
		PACKET_MAGIC,
		CMD_ACK,
		CMD_NACK,
		CMD_READY,
		CMD_DIR,
		CMD_SET_DIGITAL,
		CMD_SET_ANALOG,
		CMD_LEDRGB,
		CMD_GET,
		CMD_PUT,
		CMD_WATCH,
		
		KIND_DIGITAL,
		KIND_ANALOG,
		KIND_SERVO,
		#KIND_LEDRGB,
	]
	for i in items:
		put('%s	%02X' % (chr(i), i))
	#sys.exit()
	"""
	
	#arduino = Arduino('COM57')
	#arduino = Arduino('COM3')
	arduino = Arduino('COM56')
	
	"""
	# Protocol flood test
	#arduino.cmd(cmd, kind, index, value=None)
	time.sleep(1)
	
	put('Flooding...')
	for i in xrange(300):
		#put(str(arduino.dataQueue))
		#arduino.sendCommand_pin_none(cmd=CMD_DIR, kind='T', index=1)
		arduino.sendCommand_pin_raw(cmd=CMD_DIR, kind='T', index=1, data=bytearray([1,2,3,4,5,6,7,8,9,10,11,12]))
		#time.sleep(0.01)
	
	put('Waiting for all packets to be ACKed...')
	put(str(arduino.dataQueue))
	while arduino.waitingBytes > 0:
		put('Still waiting for %d + %d packets to be acked (%d bytes currently underway)...' % (arduino.queue.qsize(), len(arduino.dataQueue), arduino.waitingBytes))
		time.sleep(0.01)
	arduino.wait()
	put('Flooding ended cleanly.')
	
	#time.sleep(1)
	#arduino.cmd(cmd=CMD_DIR, kind='T', index=1, value=64)
	"""
	
	
	"""
	put('Blinking LED...')
	index = 0
	arduino.dir(KIND_DIGITAL, index, DIR_OUTPUT)
	for i in xrange(3):
		put('...ON...')
		arduino.setDigital(KIND_DIGITAL, index, VALUE_HIGH)
		time.sleep(0.5)
		
		put('...OFF...')
		arduino.setDigital(KIND_DIGITAL, index, VALUE_LOW)
		time.sleep(0.5)
	"""
	
	CMD_CUSTOM_PYTHON = ord('1')
	def my_onCommand(cmd, data):
		if (cmd == CMD_CUSTOM_PYTHON):
			put('my_onCommand: CMD_CUSTOM_PYTHON was sent!')
			return 1	# Stop propagation
		
		return 0	# Continue with default commands
	
	# Set callback
	arduino.onCommand = my_onCommand
	
	# Test custom command
	put('Testing custom command...')
	CMD_CUSTOM_ARDUINO = ord('0')
	arduino.sendCommand_none(CMD_CUSTOM_ARDUINO)
	time.sleep(1)
	
	
	# Test non-existent command
	put('Testing non-existent command...')
	CMD_NONEXISTENT = ord('@')
	arduino.sendCommand_none(CMD_NONEXISTENT)
	time.sleep(1)
	
	
	# Super-short
	def myCallback(value):
		put('myCallback! The pin has changed to ' + str(value) + '!')
	
	"""
	put('Testing watch on digital index 0...')
	pin = arduino.Pin(kind=KIND_DIGITAL, index=0, dir=DIR_INPUT, valueDigital=VALUE_HIGH, thresh=WATCH_ON, callback=myCallback)
	#pin = arduino.Pin(kind=KIND_DIGITAL, index=1, dir=DIR_INPUT, valueDigital=VALUE_HIGH, thresh=WATCH_ON, callback=myCallback)
	#pin = arduino.Pin(kind=KIND_DIGITAL, index=2, dir=DIR_INPUT, valueDigital=VALUE_HIGH, thresh=WATCH_ON, callback=myCallback)
	"""
	
	put('Waiting 5 seconds...')
	time.sleep(5)
	
	put('Test ended (timeout)')
	
	put('Arduino queue (empty in optimal case): ' + str(arduino.dataQueue))
	
	arduino.stop()

