#ifndef VR2
#define VR2

/*
VR2 Control
===========

Controlling the VR2 drive unit (Permobil C300 wheel chair) with an Arduino.

This sketch emulates the serial protocol used by the joystick unit of a Permobil C300 electric wheel chair.
It uses a software serial 38400 baud connection with RX and TX over the same wire - this poses some difficulties (pullups are bad here!).

The communication is very time-critical. That is why I am using a modified "CustomSoftwareSerial" library.
I needed to alter some values to make the timings perfectly. Also, I could not get it to RX correctly,
so I disabled its RX functionality (free up interrupts) and added the standard SoftwareSerial library for RX.
(It may not be able to TX at 38400,E,8,2, but when RXing it just ignores parity and superfluous stop bits - so that works).

Let there be robots!

Installation:
* Install "CustomSoftwareSerial_modified.zip" in your libraries
* Put this "vr2.h" into your sketch folder
* Include it, set the callbacks, call vr2_setup() and vr2_loop()
* Have fun!


2018-03-20 Bernhard "HotKey" Slawik


Information on the protocol
===========================



Physical
========
Now that I know the protocol (5V, 38400,E,8,2 with packets of 6 bytes with simple 0xff-X checksum), let's have some fun!
Connection is tricky (voltage-wise), because RX and TX share the same line! The pullups have to be manually set/cleared when receiving etc.
I mostly found a working combination (including mandatory timings) be trial, oscilloscope and error.

* use the WHITE wire (which is one of the 4 pins in the joystick connector; its one of the center ones)
* serial, 5V levels, 38400 baud, 1 start bit (LOW), 8 data bits, 1 parity bit (make highs EVEN), 2 stop bits (HIGH)

Connection is tricky (voltage-wise). I got it to work using an oscilloscope to carefully put the signal into OK bounds:

  WHITE (to VR2)
    |
    |
    +----------Diode >|----RX (Pin 10)
    |
    |
   10k
    |
    |
  TX (Pin 11)



What the oscilloscope shows while boot-up:
	* 26us per bit (38400 baud)
	
	* joy: ...111 0 11001010 0 11	0 00110101 0 11 111...
	       ...111 s dddddddd P SS	s dddddddd P SS 111...
		
		which transmits those bytes:
		53	MSB 01010011	LSB 11001010
		AC	MSB 10101100	LSB 00110101
	
	* ctrl: ...111 0 11001110 1 11	0 00000100 1 11	0 10000000 1 11	0 11010110 1 11 111...
	        ...111 s dddddddd P SS	s dddddddd P SS	s dddddddd P SS	s dddddddd P SS 111...
		
		which is those bytes:
		73	LSB 11001110
		20	LSB 00000100
		1	LSB 10000000
		6B	LSB 11010110
	

Packets
=======

Joystick sends this packet (each is a byte):
	cmd1	buttonState	cmd2	Y	X	checkSum
	
	cmd1:
		0x4A
	
	buttonState:
		0x00	nothing
		0x02	speed-
		0x04	speed+
		0x08	seat
		0x20	horn
		0x40	power-long-pressed
		0xC0	power
	
	cmd2:
		0xA0	standard packet
		0xA2	after beeping?
		0x01	first packet after boot
	
	X, Y:
		00	center
		00..7f	max
		ff..80	min
	
	checkSum:
		* Simple:
			checkSum = 255 - (D0 + D1 + D2 + D3 + D4)
		
		+ Example packets in decimal:
			D0	D1	D2	D3	D4	= D4
			74	0	1	1	0	= 179
			74	0	160	1	0	= 20
			74	0	160	2	0	= 19
			74	4	160	2	0	= 15
			74	0	160	1	0	= 20
			74	192	160	1	0	= 84
			74	64	160	1	0	= 212
			74	2	162	1	0	= 16
	
VR2 responds with a similar packet:
	cmd1	systemState	cmd2	batteryLed	speedLed
	
	cmd1:
		0x54
	
	systemState:
		0x80	booting?
		0xC0	occurs a few seconds after boot
		0x00	ready?
	
	cmd2:
		0xA0	normal
		0xA2	beep1 (press)
		0xA3	beep2 (maximum reached)
	
	batteryLed:
		0x51
		0x61
	
	speedLed:
		0x31
		0x51
		0x71
		0x91
		0xb1
		
		0x.3 means "inhibted by brake release lever"


Dump
====

Boot
53 AC	73 20 01 6B 
4A 00 01 01 00 B3	FE 54 80 A0 01 11 79 
4A 00 A0 01 00 14	FE 54 80 A0 01 11 79 
4A 00 A0 01 00 14	FE 54 80 A0 01 11 79 
4A 00 A0 01 00 14	FE 54 80 A0 01 11 79 

*/




//////////////////// VR2 configuration ////////////////////
//#define VR2_DEBUG // Verbosely log messages. This will interfere with binary protocols (such as pybridge)!
//#define VR2_AUTO_POWER_ON  // Auto init on start up?


// Pins for the software serials
#define VR2_CTRL_PIN_RX 10
#define VR2_CTRL_PIN_TX 11

#define DUMMY_PIN_RX 2  // Dummy pin to pass on to the TX SoftSerial implementation
#define DUMMY_PIN_TX 3  // Dummy pin to pass on to the RX SoftSerial implementation

// Internal settings of the VR2 protocol
#define VR2_BAUD 38400
#define VR2_SERIAL_SETTINGS CSERIAL_8E1 // Serial port settings
#define VR2_POWER_UP_DELAY (10+100+0)  // in milliseconds (measured: 108 milliseconds: 10ms to get from 0V to 5V, then 100ms steady 5V before first LOW transition)
#define VR2_POWER_DOWN_DELAY (20+0)  // in milliseconds (how long to keep LOW before init)
#define VR2_BYTE_DELAY (400+1000)  // delay in microseconds to bind bytes together

#define VR2_INIT_TIMEOUT 8  // Delay after initial power-up init commands in milliseconds (~8)
#define VR2_INIT_RETRIES 20 // How often to re-send the initial packet and wait for a response
#define VR2_INIT_MIN_DATA1 4  // Number of bytes to receive on init (1)
#define VR2_INIT_MIN_DATA2 6  // Number of bytes to receive on init (2)
#define VR2_RESPONSE_TIMEOUT 10 //2  // How long to wait for a response in milliseconds
#define VR2_STATE_INTERVAL_MILLIS 10  // After how many milliseconds should the state be sent again (must be < 100!)
#define VR2_FAULT_DELAY_MILLIS 10 // How long to wait after failed init retries
//const uint8_t VR2_RES_INIT[4] = {0x73, 0x20, 0x01, 0x6b};

// Commands identified so far
#define VR2_CMD_INIT 0x53
#define VR2_CMD_STATE 0x4a
#define VR2_CMD_STATE_RESPONSE 0x54

// Key codes
#define VR2_KEY_NONE 0x00
#define VR2_KEY_SLOWER 0x02
#define VR2_KEY_FASTER  0x04
#define VR2_KEY_SEAT  0x08
#define VR2_KEY_HORN  0x20
#define VR2_KEY_POWER_SHIFT 0x40
#define VR2_KEY_POWER 0xc0


// For TX using EVEN parity and 2 stop bits
#include <CustomSoftwareSerial.h>
CustomSoftwareSerial* vr2Serial;

// For RX (it just ignores parity and extra stop bit)
#include <SoftwareSerial.h>
SoftwareSerial vr2Serial_rx(VR2_CTRL_PIN_RX, DUMMY_PIN_TX);


//////////////////// the fun stuff(tm) ////////////////////
int8_t joyX, joyY;  // Current position of joystick (-127...127)
uint8_t flags;      // Yet unknown local flags (speaker)
uint8_t isConnected;    // Is there a constant stream of state information?
uint8_t reconnect;  // Automatically re-connect once connection is lost?
unsigned long timeStateLast = 0;  // Keep track on when the last state has been sent (security precaution: VR2 shuts down if no state is received)
unsigned long timePybStateLast = 0;

uint8_t state_flags1 = 0xff;
uint8_t state_flags2 = 0xff;
uint8_t state_speed = 0xff;
uint8_t state_battery = 0xff;


// Protocol buffer
#define BUFFER_SIZE 16
uint8_t buf[BUFFER_SIZE+4];
int bufLen = 0;



#ifdef VR2_DEBUG
  // Add debug implementations
  
  #define vr2_put(t) Serial.println(t)
  #define vr2_put_(t) Serial.print(t)
  #define vr2_put_hex_(t) Serial.print(t, HEX)
  #define vr2_put_float_(t) Serial.print(t, 6)

  //const char HEX_TABLE[16] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
  //char dumpBuf[BUFFER_SIZE*3+4];
  
  void vr2_put_dump(uint8_t *data, int l) {
    int i;
    vr2_put_(l); vr2_put_(F(" bytes: "));
    for (i = 0; i < l; i++) {
  
      vr2_put_hex_(data[i] >> 4);
      vr2_put_hex_(data[i] & 0x0f);
      vr2_put_(' ');
      //dumpBuf[i*3  ] = HEX_TABLE[data[i] >> 4];
      //dumpBuf[i*3+1] = HEX_TABLE[data[i] & 0x0f];
      //dumpBuf[i*3+2] = ' ';
    }
    //dumpBuf[l*3] = 0;
    //vr2_put(dumpBuf);
  }

#else
  // Remove implementations
  
  #define vr2_put(t) ;
  #define vr2_put_(t) ;
  #define vr2_put_hex_(t) ;
  #define vr2_put_dump(data, l) ;
  #define vr2_put_float_(t) ;
#endif



// Callbacks
typedef void (*t_vr2_callback)(); // We could also pass on all the data as parameters
t_vr2_callback vr2_on_state = NULL;
t_vr2_callback vr2_on_init = NULL;
t_vr2_callback vr2_on_keepAlive = NULL;
t_vr2_callback vr2_on_disconnect = NULL;
t_vr2_callback vr2_on_ready = NULL;


// Transmission
void vr2_send_data(uint8_t *data, int l) {
  //vr2_put_(F(">> TX ")); vr2_put_dump(&data[0], l);

  //vr2Serial->rx_pin_write(LOW);
  vr2Serial->tx_pin_write(HIGH);
  
  for (int i = 0; i < l; i++) {
    vr2Serial->write(data[i]);
  }
}

int vr2_receive_data(uint8_t *data, int l, int timeout) {
  // Wait for response
  unsigned long timeStart;
  unsigned long timeNow;
  int timeDelta;
  int o;
  
  timeStart = millis();
  o = 0;
  
  // Wait for data
  //digitalWrite(VR2_CTRL_PIN_RX, HIGH);
  vr2Serial->tx_pin_write(LOW);
  vr2Serial_rx.flush();
  
  while (vr2Serial_rx.available() == 0) {
    timeNow = millis();
    timeDelta = timeNow - timeStart;
    if (timeDelta > timeout) {
      vr2_put(F("Timeout!"));
      //return -1;
      o = -1;
      goto _finalize;
    }
  }
  
  while (vr2Serial_rx.available() > 0) {
    if (o < l) {
      data[o++] = (uint8_t)vr2Serial_rx.read();
      delayMicroseconds(VR2_BYTE_DELAY);  // Bind all bytes
    } else {
      vr2_put(F("Overflow!"));
      //return -1;
      break;
    }
  }

  _finalize:
  vr2Serial->tx_pin_write(HIGH);
  return o;
}


uint8_t vr2_receive_and_dump() {
  bufLen = vr2_receive_data(&buf[0], BUFFER_SIZE, VR2_RESPONSE_TIMEOUT);
  if (bufLen > 0) {
    vr2_put_(F("<< RX ")); vr2_put_dump(&buf[0], bufLen);
    return bufLen;
  } else {
    return -1;
  }
}


// Calculate VR2 checksum
uint8_t vr2_checksum(uint8_t *data, int l) {
  uint8_t check = 0xff;
  
  for (int i = 0; i < l; i++) {
    check -= data[i];
  }
  return check;
}

void vr2_send_single(uint8_t cmd) {
  buf[0] = cmd;
  buf[1] = 0xff - cmd;
  vr2_send_data(&buf[0], 2);
}



void vr2_send_state(uint8_t buttonState, uint8_t state, int8_t stateY, int8_t stateX) {
  buf[0] = VR2_CMD_STATE;
  buf[1] = buttonState;
  buf[2] = state; // 0xa0 + speaker
  buf[3] = (uint8_t)stateY;
  buf[4] = (uint8_t)stateX;
  buf[5] = vr2_checksum(&buf[0], 5);
  vr2_send_data(&buf[0], 6);

  timeStateLast = millis();
}



// Perform init sequence
void vr2_init() {
  int i;
  unsigned long timeNow;
  unsigned long timeStart;
  int timeDelta;
  int retries = VR2_INIT_RETRIES;

  isConnected = 0;
  state_flags1 = 0xff;
  state_flags2 = 0xff;
  state_speed = 0xff;
  state_battery = 0xff;
  
  vr2_put(F("vr2_init..."));
  
  // Callback
  if (vr2_on_init != NULL)
    vr2_on_init();

  vr2_put(F("vr2_init: LOW..."));
  
  digitalWrite(VR2_CTRL_PIN_RX, LOW);
  //vr2Serial->rx_pin_write(LOW);
  vr2Serial->tx_pin_write(LOW);

  //delay(VR2_POWER_DOWN_DELAY);
  for (i = 0; i < 10; i++) {
    if (vr2_on_keepAlive != NULL)
      vr2_on_keepAlive();
    delay(VR2_POWER_DOWN_DELAY / 10);
  }

  // Power cycle
  vr2_put(F("vr2_init: HIGH..."));
  digitalWrite(VR2_CTRL_PIN_RX, HIGH);
  vr2Serial->tx_pin_write(HIGH);
  //digitalWrite(VR2_CTRL_PIN_RX, LOW);

  
  //delay(VR2_POWER_UP_DELAY);
  for (i = 0; i < 10; i++) {
    if (vr2_on_keepAlive != NULL)
      vr2_on_keepAlive();
    delay(VR2_POWER_UP_DELAY / 10);
  }

  vr2_put(F("vr2_init: serial..."));
  
  vr2Serial->begin(VR2_BAUD, VR2_SERIAL_SETTINGS);
  
  digitalWrite(VR2_CTRL_PIN_RX, LOW); // Need it low!
  
  while (retries > 0) {

    // Keep alive (for tasks that need constant attention, like some SoftwareSerial implementations)
    if (vr2_on_keepAlive != NULL)
      vr2_on_keepAlive();
    
    
    //vr2Serial->rx_pin_write(LOW);
    //digitalWrite(VR2_CTRL_PIN_RX, LOW);
    
    // Send init sequence (power button)
    vr2_put(F(">> TX VR2_CMD_INIT"));
    vr2_send_single(VR2_CMD_INIT);  // sends 0x53 0xac
    
    vr2_put_(F("<< RX VR2_RES_INIT: "));
    bufLen = vr2_receive_data(&buf[0], BUFFER_SIZE, VR2_INIT_TIMEOUT);
    if (bufLen <= 0) {
      vr2_put(F("Nothing (1)"));
      goto retry;
    }
    if (bufLen < VR2_INIT_MIN_DATA1) {
      vr2_put(F("Not enough (1)"));
      goto retry;
    }
    vr2_put_dump(&buf[0], bufLen);
    
    
    vr2_put(F(">> TX VR2_CMD_STATE"));
    vr2_send_state(0x00, 0x01, 0x01, 0x00);
    
    vr2_put_(F("<< RX VR2_RES_STATE: "));
    bufLen = vr2_receive_data(&buf[0], BUFFER_SIZE, VR2_INIT_TIMEOUT);
    if (bufLen <= 0) {
      vr2_put(F("Nothing (2)"));
      goto retry;
    }
    if (bufLen < VR2_INIT_MIN_DATA2) {
      vr2_put(F("Not enough (2)"));
      goto retry;
    }
    vr2_put_dump(&buf[0], bufLen);
    
    
    // All tests have succeeded
    
    
    vr2_put(F("Ready!"));
    isConnected = 1;
    timeStateLast = millis();

    if (vr2_on_ready != NULL)
      vr2_on_ready();
    
    return;

retry:
    vr2_put(F("Retrying..."));
    retries--;
  }
  
  vr2_put(F("Giving up."));
  vr2Serial->end();
  vr2Serial->tx_pin_write(LOW);

  //vr2Serial->rx_pin_write(LOW);
  //digitalWrite(VR2_CTRL_PIN_RX, LOW);
  delay(VR2_FAULT_DELAY_MILLIS);
}
uint8_t vr2_handle_state_response(uint8_t *data) {
  if (data[0] != VR2_CMD_STATE_RESPONSE) return 0;
  if (vr2_checksum(&data[0], 5) != data[5]) return 0;
  
  //vr2_put_(F("VR2_STATE_RESPONSE: "));
  //vr2_put_dump(&data[0], 6);
  
  byte has_changed = 0;
  
  if (state_flags1 != data[1]) has_changed++;
  state_flags1 = data[1]; // 0x00 = ready, 0x80 = booting done, 0xc0 = error
  
  if (state_flags2 != data[2]) has_changed++;
  state_flags2 = data[2]; // 0xa0 / 0xa2/0xa3 Speaker?
  
  if (state_battery != data[3]) has_changed++;
  state_battery = data[3];  // 0x61
  
  if (state_speed != data[4]) has_changed++;
  state_speed = data[4];  // 0x31, 0x51, 0x71, 0x91, 0xb1
  
  
  // Parse
  /*
normal:
  flags=0, beep=A0, batt=61, speed=31
  flags=0, beep=A0, batt=61, speed=31
  flags=0, beep=A0, batt=61, speed=31
  flags=0, beep=A0, batt=61, speed=31

error:
  flags=80, beep=A0, batt=73, speed=10
  flags=80, beep=A0, batt=73, speed=10
  flags=80, beep=A0, batt=73, speed=10
  */
  
  if (state_flags1 & 0x80) {
    // Error condition: Can occur shortly at boot or constantly when there is something wrong and drive is inhibited
    /*
    #ifdef INTERFACE_PYBRIDGE
      pyb_send_info(INFO_ERROR);
    #endif
    */

    //@TODO: Send ERROR and reset connection!
    
  } else {
    
    // Normal operation
    
  }
  
  vr2_put_(F("f1=")); vr2_put_hex_(state_flags1);
  vr2_put_(F("\tf2=")); vr2_put_hex_(state_flags2);
  vr2_put_(F("\tbat=")); vr2_put_hex_(state_battery);
  vr2_put_(F("\tspeed=")); vr2_put_hex_(state_speed);
  vr2_put();
  
  
  if (has_changed > 0) {
    
    // Callback
    if (vr2_on_state != NULL)
      vr2_on_state();
    
  }
  
  return 1;
}

// Receive, clean, check and parse the state response from VR2
uint8_t vr2_receive_and_handle_state_response() {
  uint8_t valid = 0;
  // Receive answer
  //bufLen = vr2_receive_and_dump();
  bufLen = vr2_receive_data(&buf[0], BUFFER_SIZE, VR2_RESPONSE_TIMEOUT);

  // Check for valid packet (strip crap by trying all possibilities until checksum fits)
  if (bufLen <= 0) {
    // Something went wrong
    vr2_put(F("VR2 connection lost!"));
    
    isConnected = 0;
    state_flags1 = 0xff;
    state_flags2 = 0xff;
    state_speed = 0xff;
    state_battery = 0xff;

    if (vr2_on_disconnect != NULL)
      vr2_on_disconnect();
    
    valid = 0;
  
  } else
  if (bufLen >= 6) {
    // Enough data
    for (int i = 0; i <= bufLen-6; i++) {
      //if (buf[i] != VR2_CMD_STATE_RESPONSE) continue;
      //if (vr2_checksum(&buf[i], 5) != buf[i+5]) continue;
      
      if (vr2_handle_state_response(&buf[i]) == 1) {
        // Found a parsable chunk
        valid = 1;
        return 1;
        break;
      }
    }
  }

  if ((valid == 0) && (bufLen > 0)) {
    // Unhandled data
    
    if ((bufLen == 1) && (buf[0] == 0)) {
      // Skip null byte data
    } else {
      vr2_put_(F("Erroneous: ")); vr2_put_dump(&buf[0], bufLen);  // Output anything else
    }
  }

  return valid;
}


void vr2_press_key(uint8_t key) {
  vr2_send_state(key, flags, joyX, joyY);
  vr2_receive_and_handle_state_response();
}

void vr2_set_joystick(int8_t x, int8_t y) {
  joyX = x;
  joyY = y;
  
  // Send right away?
  //vr2_send_state(VR2_KEY_NONE, flags, joyX, joyY);
  //vr2_receive_and_handle_state_response();
}


void vr2_power_off() {
  vr2_put(F("Power off"));

  joyX = 0;
  joyY = 1;
  
  vr2_send_state(VR2_KEY_POWER, flags, joyX, joyY);
  vr2_receive_and_handle_state_response();
  
  vr2Serial->end();
  vr2Serial->rx_pin_write(LOW);
  vr2Serial->tx_pin_write(LOW);

  isConnected = 0;
  state_flags1 = 0xff;
  state_flags2 = 0xff;
  state_speed = 0xff;
  state_battery = 0xff;

  if (vr2_on_state != NULL)
    vr2_on_state();
  
  // Prevent re-connection
  reconnect = 0;
}



void vr2_setup() {
  // Setup VR2...
  // Using https://github.com/ledongthuc/CustomSoftwareSerial
  //vr2Serial = new CustomSoftwareSerial(VR2_CTRL_PIN_RX, VR2_CTRL_PIN_TX);
  vr2Serial = new CustomSoftwareSerial(DUMMY_PIN_RX, VR2_CTRL_PIN_TX);
  //vr2Serial->begin(VR2_BAUD, VR2_SERIAL_SETTINGS);
  //vr2Serial->end();
  vr2Serial->tx_pin_write(LOW);
  //vr2Serial->rx_pin_write(LOW);
  
  vr2Serial_rx.begin(VR2_BAUD);
  digitalWrite(VR2_CTRL_PIN_RX, LOW); // Need it low!
  
  
  // Create sane state
  isConnected = 0;
  state_flags1 = 0xff;
  state_flags2 = 0xff;
  state_speed = 0xff;
  state_battery = 0xff;
  
  reconnect = 0;
  flags = 0xa0;
  joyX = 0;
  joyY = 0;

  
  #ifdef VR2_AUTO_POWER_ON
    // Connect...
    reconnect = 1;
    vr2_init();
  #endif
}

void vr2_loop() {
  unsigned long timeNow = millis();
  int timeDelta;
  
  if ((isConnected == 0) && (reconnect == 1)) {
    // Not connected? Auto-connect!
    
    vr2_put(F("Auto-reconnecting..."));
    vr2_init();
  }
  
  // Keep VR2 connection alive (has a watchdog that WILL shut down the connection after 50ms or so)
  if (isConnected == 1) {
    timeDelta = timeNow - timeStateLast;
    if ((timeDelta > VR2_STATE_INTERVAL_MILLIS) || (timeDelta < 0)) {
      
      // Send current state to VR2
      vr2_send_state(VR2_KEY_NONE, flags, joyY, joyX);
      vr2_receive_and_handle_state_response();
      
      timeStateLast = timeNow;
    }
  }
  
  
}


#endif

