#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import sys
sys.path.append('../pybridge')
sys.path.append('lib/pybridge')
sys.path.append('../lib/pybridge/python')
from arduino import *

ARDUINO_SERIAL_PORT = 'COM56'	# SiQo
#ARDUINO_SERIAL_PORT = 'COM4'	# SuQu2
#ARDUINO_SERIAL_BAUD = 115200
ARDUINO_SERIAL_BAUD = 57600

#import math	# for atan2 / magnetic heading

USE_GPS = True
USE_COMPASS = True
USE_NEOPIXEL = True

# Custom commands
CMD_VR2_INFO = ord('i')
CMD_VR2_STATE = ord('s')
CMD_VR2_POWER_OFF = 0
CMD_VR2_POWER_ON = 1
CMD_VR2_JOY = ord('j')
CMD_VR2_KEY = ord('k')
CMD_VR2_STOP = ord('.')
#CMD_VR2_SPEED_FASTER = ord('+')
#CMD_VR2_SPEED_SLOWER = ord('-')

if USE_GPS:
	CMD_LOCATION = ord('L')

if USE_COMPASS:
	CMD_MAGNET = ord('M')
	MAX_MAGNET_UT = 1000.0	# Must be kept in-sync with Arduino sketch
	CMD_HEADING = ord('H')

if USE_NEOPIXEL:
	PIXELS = 8
	CMD_NEOPIXEL_FILL = ord('P')
	CMD_NEOPIXEL_GRADIENT = ord('G')
	CMD_NEOPIXEL_RAW = ord('W')

# Infos (CMD_VR2_INFO)
INFO_INIT = 0x00
INFO_READY = 0x01
INFO_DISCONNECTED = 0x02
INFO_ERROR = 0x03
INFOS = ['INIT', 'READY', 'DISCONNECTED', 'ERROR']

# Key codes (CMD_VR2_KEY)
VR2_KEY_NONE = 0x00
VR2_KEY_SLOWER = 0x02
VR2_KEY_FASTER = 0x04
VR2_KEY_SEAT = 0x08
VR2_KEY_HORN = 0x20
VR2_KEY_POWER_SHIFT = 0x40
VR2_KEY_POWER = 0xc0

# Limits
VR2_MAX_JOYSTICK = 100
VR2_MAX_GEAR = 4	# 0=3x, 1=5x, 2=7x, 3=9x, 4=Bx = 0..4
VR2_MAX_BATTERY = 11	# Used to calculate battery percentage; values are 0x...Fx?	(10 LEDs 0x6x=5 lit = 0...11)

# Timings
VR2_POWER_ON_DELAY = 2.0	# Delay in seconds to wait after connected
VR2_GEAR_DELAY = 0.5	# Delay after setting gear/speed via button press


def put(txt):
	print('vr2:\t' + str(txt))

# Converters from/to signed int8
def byte_to_sint8(v):
	return v if (v < 0x80) else v-0x100

def sint8_to_byte(v):
	return int(min(0x7f, v)) if (v >= 0) else int(max(-127, v) + 0x100)


if USE_GPS:
	def uint32_to_deg(i):
		f = i * (360.0 / 0x100000000L)
		#put('uint32_to_deg: %d = 0x%08X = %f' % (i, i, f))
		return f

	def uint16_to_deg(i):
		f = i * (360.0 / 0x10000)
		#put('uint16_to_deg: %d = 0x%04X = %f' % (i, i, f))
		return f

if USE_COMPASS:
	def uint16_to_ut(i):
		t = ((2.0 * float(i) / float(0xffff)) - 1.0) * MAX_MAGNET_UT
		return t


class VR2:
	"Class (based on pybridge) to control the functions of a VR2 drive unit (e.g. Permobil C300 wheel chair)"
	
	def __init__(self, serial_port, serial_baud=ARDUINO_SERIAL_BAUD, autostart=True):
		
		self.serial_port = serial_port
		self.serial_baud = serial_baud
		
		# The fun stuff (parsed and usable values)
		self.is_connected = False
		self.is_ready = False
		self.is_inhibited = False
		
		self.battery = 0.0	# 0.0 ... 1.0
		self.gear = 0	# 0..VR2_MAX_GEAR
		self.speaker = 0	# 0..3 for different beeps
		self.joy = [0.0, 0.0]	# -1.0 ... +1.0
		
		
		# GPS Compass
		if USE_GPS:
			self.pos_geo = [0.0, 0.0, 0.0]	# lat, lon [deg] alt [m]
			self.on_pos_geo = None
		
		if USE_COMPASS:
			self.magnet = [0.0, 0.0, 0.0]	# x, y, z [uT]
			self.on_magnet = None
			self.heading = 0.0
			self.on_heading = None
		
		
		# Local copy of the VR2 state vars as reported (unparsed)
		self.state_is_connected = 0
		self.state_flags1 = 0
		self.state_flags2 = 0
		self.state_speed = 0
		self.state_battery = 0
		self.state_joy_x = 0
		self.state_joy_y = 0
		
		self.on_status = None
		
		# pybridge to communicate
		self.arduino = None	# Create it before opening (in case an exception triggers __del__())
		
		if autostart:
			self.start()
	
	def start(self):
		put('Connecting to Arduino on port "%s" at %d baud...' % (self.serial_port, self.serial_baud))
		self.arduino = Arduino(port=self.serial_port, baud=self.serial_baud, onCommand=self.on_pyb_command)
	
	def __del__(self):
		self.kill()
	
	def kill(self):
		if (self.arduino is not None):
			# Stop moving
			self.stop()
			
		
		# Make sure no one accesses us
		self.is_ready = False
		self.is_connected = False
		
		if (self.arduino is not None):
			# Disconnect
			self.power_off()
			
			# Wait
			time.sleep(1)
			
			# Kill connection to arduino
			self.arduino.stop()
		#
	
	
	def on_pyb_command(self, cmd, data):
		#put('arduino_on_command...')
		#for b in data: put('%02X' % (b))
		
		if (cmd == CMD_VR2_INFO):
			# Simple info report
			
			if (len(data) < 1): return 1	# Stop propagation
			
			i = int(data[0])
			info_text = INFOS[i] if i < len(INFOS) else '?'
			put('Info: 0x%02X = %s' % (i, info_text))
			
			return 1	# Stop propagation
		
		elif (cmd == CMD_VR2_STATE):
			# State report
			
			if (len(data) < 7): return 1	# Stop propagation
			
			has_changed = False
			
			# Just copy to local copy
			has_changed = has_changed or (self.state_is_connected != data[0])
			self.state_is_connected = data[0]
			
			has_changed = has_changed or (self.state_flags1 != data[1])
			self.state_flags1 = data[1]
			
			has_changed = has_changed or (self.state_flags2 != data[2])
			self.state_flags2 = data[2]	# Contains some buzzer info on lowest 2 bits
			
			has_changed = has_changed or (self.state_speed != data[3])
			self.state_speed = data[3]
			
			has_changed = has_changed or (self.state_battery != data[4])
			self.state_battery = data[4]	# 10 LEDs: 5 lit = 0x61
			
			# Not part of VR2 state, but it is nice to know where Arduino thinks we are
			self.state_joy_x = byte_to_sint8(data[5])
			self.state_joy_y = byte_to_sint8(data[6])
			
			put('State: is_connected=%d	flags1=0x%02X	flags2=0x%02X	speed=0x%02X	battery=0x%02X	x=%d	y=%d' % (self.state_is_connected, self.state_flags1, self.state_flags2, self.state_speed, self.state_battery, self.state_joy_x, self.state_joy_y))
			
			
			# Parse states to usable values
			
			# Does the Arduino report "connected"?
			self.is_connected = True if (self.state_is_connected > 0) else False
			
			# Parse values according to state
			if self.is_connected:
				
				# Connected. Does VR2 report an error condition?
				if (self.state_flags1 & 0x80):
					# Error condition!
					self.is_ready = False
				
				else:
					# Default condition (connected, no error)
					self.is_ready = True
					
					# Parse state_battery: 10 LEDs total; 5 lit = 0x61
					b = (self.state_battery >> 4)	# 0..B (0..11)
					self.battery = float(b) / VR2_MAX_BATTERY
					#self.battery = min(1.0, max(0.0, self.battery))	# Clip
					
					# Parse state_speed: 0x31,0x51,0x71,0x91,0xB1
					s = (self.state_speed >> 4)	# 3,5,7,9,B in steps of 2
					self.gear = (s - 3) // 2	# 0..4
					#self.gear = min(VR2_MAX_GEAR, max(0, self.gear))	# Clip
					
					# If an inhibitor sensor is triggered (e.g. manual brake-release), the low nibble of state_speed becomes 0x02 instead of 0x01
					self.is_inhibited = (self.state_speed & 0x02) > 0
					
					self.speaker = (self.state_flags2 & 0x0f)	# 0..3 for different beeps
			
			else:
				
				# Not connected = not ready
				self.is_ready = False
			
			if has_changed:
				if not self.on_status is None:
					self.on_status()
			
			return 1	# Stop propagation
			
		
		
		elif (USE_GPS) and (cmd == CMD_LOCATION):
			lat = uint32_to_deg((data[0] * 0x01000000L) + (data[1] * 0x00010000L) + (data[2] * 0x00000100L) + data[3])
			lon = uint32_to_deg((data[4] * 0x01000000L) + (data[5] * 0x00010000L) + (data[6] * 0x00000100L) + data[7])
			alt = (data[8] * 0x0100) + data[9] - 0x8000
			self.pos_geo = (lat, lon, alt)
			
			if not self.on_pos_geo is None:
				self.on_pos_geo(self.pos_geo)
			
			return 1	# Stop propagation
		
		elif (USE_COMPASS) and (cmd == CMD_MAGNET):
			# Handle raw magnetic values
			mag_x = uint16_to_ut((data[0] * 0x0100) + data[1])
			mag_y = uint16_to_ut((data[2] * 0x0100) + data[3])
			mag_z = uint16_to_ut((data[4] * 0x0100) + data[5])
			self.magnet[:] = mag_x, mag_y, mag_z
			
			#self.heading = math.atan2(mag_z, mag_x) * 180.0 / math.pi
			
			if not self.on_magnet is None:
				self.on_magnet(self.magnet)
			
			return 1	# Stop propagation
		
		elif (USE_COMPASS) and (cmd == CMD_HEADING):
			# This is "subjective" to the Arduino sketch's hard-coded magnetic declination
			heading = uint16_to_deg((data[0] * 0x0100) + data[1])
			self.heading = heading
			
			if not self.on_heading is None:
				self.on_heading(heading)
			
			return 1	# Stop propagation
		
		put('on_pyb_command: Nothing to handle (%02X)' % (cmd))
		return 0	# Continue with default commands
	
	
	def pyb_send_command(self, cmd, data=None):
		self.arduino.sendCommand_data(cmd, data)
		
	
	def blink(self):
		put('Blinking LED...')
		index = 0
		self.arduino.dir(KIND_DIGITAL, index, DIR_OUTPUT)
		for i in xrange(3):
			put('...ON...')
			self.arduino.setDigital(KIND_DIGITAL, index, VALUE_HIGH)
			time.sleep(0.5)
			
			put('...OFF...')
			self.arduino.setDigital(KIND_DIGITAL, index, VALUE_LOW)
			time.sleep(0.5)
	
	def power_on(self, wait=True):
		put('Powering ON...')
		
		self.is_connected = False	# Assume we are not connected, yet
		self.is_ready = False	# To be sure: signal "not ready"
		
		self.arduino.sendCommand_none(CMD_VR2_POWER_ON)
		if wait:
			put('Connecting to VR2...')
			while (not self.is_ready):
				time.sleep(0.2)
			put('Waiting...')
			time.sleep(VR2_POWER_ON_DELAY)
			
			# Check again after the delay
			if (not self.is_ready):
				put('Power on failed (went disconnected again)')
				return False
			
		put('Powered ON!')
		return True
	
	def power_off(self):
		put('Power OFF...')
		self.is_ready = False	# Signal that we might not be ready soon...
		self.pyb_send_command(CMD_VR2_POWER_OFF)
	
	def set_joystick(self, x, y):
		# Clip
		x = min(1.0, max(-1.0, x))
		y = min(1.0, max(-1.0, y))
		self.joy[0] = x
		self.joy[1] = y
		
		if (not self.is_ready):
			put('Not setting joystick, because not ready!')
			return False
			
		# Convert
		x = int(max(-VR2_MAX_JOYSTICK, min(VR2_MAX_JOYSTICK, x * VR2_MAX_JOYSTICK)))
		y = int(max(-VR2_MAX_JOYSTICK, min(VR2_MAX_JOYSTICK, y * VR2_MAX_JOYSTICK)))
		
		put('Setting joystick to (sint8) %d/%d' % (x,y))
		
		self.pyb_send_command(CMD_VR2_JOY, [sint8_to_byte(x), sint8_to_byte(y)])
		return True
	
	def press_key(self, k):
		"Press a key on the control panel"
		if (not self.is_ready):
			put('Not pressing key, because not ready!')
			return False
		
		put('Pressing key 0x%02X' % (k))
		self.pyb_send_command(CMD_VR2_KEY, [k])
	
	def stop(self):
		"Set joystick to neutral. Stop."
		if (not self.is_ready):
			put('Not stopping, because not ready!')
			return False
		
		put('Stopping')
		self.joy = [0.0, 0.0]
		self.pyb_send_command(CMD_VR2_STOP)
		return True
	
	def set_gear(self, gear):
		"Set the speed level"
		gear = max(0, min(VR2_MAX_GEAR, gear))
		put('Setting gear from %d to %d/%d...' % (self.gear, gear, VR2_MAX_GEAR))
		
		# Now press the buttons until we reached the desired gear
		while (self.gear != gear):
			
			if (not self.is_ready):
				put('Gear not set, because not ready!')
				return False
			
			if (self.gear < gear):
				#put('Faster...')
				self.press_key(VR2_KEY_FASTER)
			elif (self.gear > gear):
				#put('Slower...')
				self.press_key(VR2_KEY_SLOWER)
			time.sleep(VR2_GEAR_DELAY)
		
		put('Gear is set to %d of %d' % (self.gear, VR2_MAX_GEAR))
		return True
	
	"""
	# Just do a "set_speed" (just one command) or just a "press VR2 key" command
	def speed_faster(self):
		self.pyb_send_command(CMD_VR2_SPEED_FASTER)
		
	def speed_slower(self):
		self.pyb_send_command(CMD_VR2_SPEED_SLOWER)
		
	"""
	
	### NeoPixel stuff
	def neopixel_fill(self, col, count=PIXELS):
		#self.pyb_send_command(CMD_NEOPIXEL_FILL, [col[0], col[1], col[2]])
		#self.pyb_send_command(CMD_NEOPIXEL_FILL, [col[2], col[1], col[0]])	# BGR conversion
		self.pyb_send_command(CMD_NEOPIXEL_FILL, [col[2], col[1], col[0], count])	# BGR conversion
	
	def neopixel_raw(self, pixels):
		"Send raw RGB values for each pixel"
		data = []
		for p in pixels:
			#data.append(p[0]);data.append(p[1]);data.append(p[2])
			data.append(p[2]);data.append(p[1]);data.append(p[0])	# BGR conversion
		self.pyb_send_command(CMD_NEOPIXEL_RAW, data)
	
	def neopixel_gradient(self, grads):
		"tuples of <constant_length, new_color, gradient_length>"
		data = []
		for g in grads:
			data.append(g[0])	# Constant length (initially for BLACK)
			data.append(g[1][2])	# new color B
			data.append(g[1][1])	# new color G
			data.append(g[1][0])	# new color R
			data.append(g[2])	# gradient run length
		self.pyb_send_command(CMD_NEOPIXEL_GRADIENT, data)


if __name__ == '__main__':
	put('VR2 test using pybridge')
	
	vr2 = VR2(serial_port=ARDUINO_SERIAL_PORT, serial_baud=ARDUINO_SERIAL_BAUD)
	
	if USE_GPS:
		def my_on_pos_geo(geo):
			put('on_pos_geo: %3.6f, %3.6f %0.2f' % (geo[0], geo[1], geo[2]))
		vr2.on_pos_geo = my_on_pos_geo
	
	if USE_COMPASS:
		def my_on_magnet(magnet):
			put('on_magnet [uT]: %s' % (str(magnet)))
		vr2.on_magnet = my_on_magnet
		
		def my_on_heading(heading):
			put('on_heading: %3.1f deg' % (heading))
		vr2.on_heading = my_on_heading
	
	for i in xrange(3):
		vr2.neopixel_fill((0x00, 0x10, 0x00))
		time.sleep(0.25)
		vr2.neopixel_fill((0x00, 0x00, 0x00))
		time.sleep(0.25)
	
	vr2.neopixel_raw([
		(0x00, 0x00, 0x00),
		(0x10, 0x00, 0x00),
		(0x20, 0x00, 0x00),
		(0x00, 0x10, 0x00),
		(0x00, 0x20, 0x00),
		(0x00, 0x00, 0x10),
		(0x00, 0x00, 0x20),
		(0x10, 0x10, 0x10),
	])
	time.sleep(1.0)
	
	
	
	vr2.neopixel_gradient([
		(3,	(0x40, 0x00, 0x00),	3),
		(3,	(0x00, 0x40, 0x00),	3),
		(3,	(0x00, 0x00, 0x40),	3),
	])
	time.sleep(1.0)
	"""
	vr2.neopixel_gradient([
		(0,	(0x40, 0x00, 0x00),	4),
		(0,	(0x00, 0x40, 0x00),	4),
	])
	time.sleep(1.0)
	vr2.neopixel_gradient([
		(2,	(0x40, 0x00, 0x00),	3),
		(8,	(0x00, 0x00, 0x40),	8),
	])
	time.sleep(1.0)
	"""
	
	
	"""
	#vr2.blink()
	put('Powering on...')
	vr2.power_on(wait=True)
	
	put('VR2 is powered on!')
	time.sleep(1)
	
	
	put('Setting Gear to 2...')
	vr2.set_gear(2)
	time.sleep(1)
	
	put('Playing with the joystick...')
	vr2.set_joystick(0, 0.2)
	time.sleep(3)
	
	vr2.set_joystick(0, 0.4)
	time.sleep(3)
	
	put('Faster gear...')
	vr2.set_gear(3)
	
	put('Curve...')
	vr2.set_joystick(-0.4, 0.4)
	time.sleep(3)
	
	put('Slower gear...')
	vr2.set_gear(2)
	
	put('Coming to a rest...')
	vr2.set_joystick(0, 0.2)
	time.sleep(3)
	
	put('Stopping...')
	vr2.stop()
	time.sleep(3)
	"""
	
	put('Safe power down...')
	vr2.stop()
	time.sleep(1)
	
	vr2.power_off()
	time.sleep(1)
	
	
	put('All in all %d packets have not been ACKed: %s' % (len(vr2.arduino.dataQueue), str(vr2.arduino.dataQueue)))
	put('End.')
	