#ifndef PYBRIDGE
#define PYBRIDGE

/*
 * pybridge
 * Arduino Part
 * 
 * This is something like Firmata, but I want to write my own stuff. Sorry.
 * It takes care of sequence numbers, re-sending, chacksums.
 * So it is a low level transport layer.
 * 
 * (C) 2015-2018 Bernhard Slawik
 * 
 * 
 * Protocol:
 *   * Commands are sent using a simple package structure that takes care of checksums, flow control etc.
 *   * Packet layout:
 *      MAGIC  CMD  SEQ [LEN  PAYLOAD...] CHECK
 *      |      |    |    |    |           |
 *      |      |    |    |    |           |__ BYTE: Checksum. Simply "0xFF XOR <all bytes after MAGIC>"
 *      |      |    |    |    |
 *      |      |    |    |    |__ BYTEARRAY: Payload bytes of variable length.
 *      |      |    |    |
 *      |      |    |    |__ BYTE: Length of payload
 *      |      |    |
 *      |      |    |__ BYTE: Sequence number
 *      |      |
 *      |      |__ BYTE: Command
 *      |      
 *      |
 *      |__ BYTE: Start byte (PACKET_MAGIC) for synchronization, usually "!"
 * 
 *   * Commands are:
 *     * CMD_ACK SEQ
 *         Acknowledge the correct reception of packet with sequence number SEQ
 *     * CMD_NACK SEQ
 *         Problems receiving package with sequence number SEQ (checksums differ, length strange, ...)
 *
 *     * CMD_xxx KIND INDEX VALUE
 *         Commands to actually do something with the Arduino pins.
 *         * KIND
 *             Which kind of pin group is addressed (digital, analog, servo, ...). Remember: You can set an ANALOG value to a DIGITAL PWM pin and vice versa, so we need to be able to do e.g. "SET_ANALOG DIGITAL #0 to 0x200"
 *         * INDEX
 *             The pin index (see xxx_pin[])
 *         * VALUE
 *             Depending on the command (analog/digital) this can be a BYTE (CMD_SET_DIGITAL) or WORD (CMD_SET_ANALOG)
 * 
 *     * CMD_WATCH KIND INDEX THRESHOLD
 *         Watch for changes on the given pin. A CMD_PUT response is triggered when the pin value changed above the given THRESHOLD (word). A THRESHOLD of 0000 means "un-watch".
 * 
 *     * CMD_LEDRGB SUBCOMMAND VALUE...
 *         Commands to mess with WS2812 RGB LEDs. There are SUBCOMMANDs to clear, fill, set one pixel, set all pixels. Layout should be obvious (see implementation).
 *
 *
 * TODO:
 * + "Kind Command rest..." instead of "Command Kind Index = Value"
 * 
 */

// Setup (define those in your code before including this file)

// Select functionality
//#define USE_DIGITALS
//#define USE_ANALOGS
//#define USE_SERVOS
//#define USE_LEDRGB
//#define USE_CUSTOM_COMMANDS // Enable pybridge_on_command callback

// This creates a lot of serial debug messages. This also increases the global variable usage to insane levels...
//#define DEBUG_VERBOSE


// Select which serials to use
//#define SerialDebug Serial
//#define SerialCommand Serial

#define pyb_put(t) SerialDebug.println(t)
#define pyb_put_(t) SerialDebug.print(t)


#ifdef USE_CUSTOM_COMMANDS
  typedef byte (*t_pybridge_on_command)(byte cmd, byte len, byte *data);
  t_pybridge_on_command pybridge_on_command = NULL;
#endif


#define LF "\n"

#define COMMAND_TIMEOUT_MILLIS 200
#define COMMAND_MAX_INPUT 64	// See HardwareSerial.cpp:SERIAL_BUFFER_SIZE
//#define COMMAND_MAX_INPUT 256
//#define COMMAND_MAX_INPUT 128

#define COMMAND_DIGITS 2
#define COMMAND_MAX 0xff

#define PACKET_MAGIC '!'

#define CMD_ACK 'A'
#define CMD_NACK 'N'
#define CMD_UNKNOWN '?'
#define CMD_ERROR 'X'
#define CMD_READY 'R'

#define CMD_GET 'g'
#define CMD_PUT 'p'
#define CMD_WATCH 'w'
#define CMD_DIR 'r'
#define CMD_SET_DIGITAL 'd'
#define CMD_SET_ANALOG 'a'
#define CMD_LEDRGB 'l'

#define KIND_DIGITAL 'D'
#define KIND_ANALOG 'A'
#define KIND_SERVO 'S'
//#define KIND_LEDRGB 'L'

#define DIR_INPUT 'I'
#define DIR_OUTPUT 'O'

#define ERROR_LENGTH 1  // Error about payload size
#define ERROR_SEQ_DELTA 2 // Acked and last acked sequence numbers differ more than 1
#define ERROR_OBSOLETE 3  // NACK requested a packet that is not available any more

// Protocol states
// PACKET_MAGIC  SEQ LEN  DATA... CHECK
#define OFS_START 0
#define OFS_CMD 1
#define OFS_SEQ 2
#define OFS_LEN 3
#define OFS_DATA 4
#define OFS_CHECK 5

// Receiver states
byte rPacketState = OFS_START;
byte rPacketCmd = 0;
byte rPacketSeq = 0;
byte rPacketPayloadLen = 0;
byte rPacketPayloadLenCurrent = 0;
byte rPacketPayload[COMMAND_MAX_INPUT];
byte rPacketCheckSupposed = 0;
byte rPacketCheckActual = 0;

unsigned long rPacketStartMillis;  // Keep track when last packet has begun so we can catch timeouts and re-set
byte rPacketSeqLastAcked = 0;


// Sender states
byte sPacketSeqLastSent = 0;

// Last transmission (for re-transmission)
byte sPacketCmdLast;
byte sPacketPayloadLenLast;
byte *sPacketPayloadLast;




/*
	Hardware pins are mapped to "logical" pins to abstract from the actual microcontroller hardware.
	This allows to "set servo #0" or "set digital #1" instead of taking care which pins are capable of what in the actual hardware.
	
	For example:
		* Arduino pins 0 and 1 are usually RX/TX, so they can not be used for controlling stuff. Digital indices would therefore have to start from "2" which may cause some confusion. So with this implementation we can simply enumerate them from 0..DIGITALS-1
		* Some pins are PWM capable. We should reserve them for special uses and exclude them from the list of "standard digital IO pins".
		* Some libraries depend on certain pin layouts. Those reserved pins can be excluded so we don't interfere with them.
*/


#ifdef USE_DIGITALS
  #if defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
    // MEGA: Max 54 Digitals
    #define DIGITALS 12
    int  digitals_pin[DIGITALS] = {22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33};
    byte digitals_dir[DIGITALS] = {INPUT, INPUT, INPUT, INPUT, INPUT, INPUT, INPUT, INPUT, INPUT, INPUT, INPUT, INPUT};
    byte digitals_value[DIGITALS] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
    byte digitals_watch[DIGITALS] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  #endif

  #if defined(__AVR_ATmega328P__) || defined(__AVR_ATmega168__)
    // UNO etc.
    #define DIGITALS 2
    int  digitals_pin[DIGITALS] = {13, 12};
    byte digitals_dir[DIGITALS] = {INPUT, INPUT};
    byte digitals_value[DIGITALS] = {0xff, 0xff};
    byte digitals_watch[DIGITALS] = {0, 0};
  #endif
  
#endif


#ifdef USE_ANALOGS
  //#if defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
    // Max 16 Analogs on MEGA
    #define ANALOGS 4
    int  analogs_pin[ANALOGS] = {A0, A1, A2, A3};
    byte analogs_dir[ANALOGS] = {INPUT, INPUT, INPUT, INPUT};
    int  analogs_value[ANALOGS] = {-1, -1, -1, -1};
    byte analogs_watch[ANALOGS] = {0, 0, 0, 0};
  //#endif
  
#endif

#ifdef USE_SERVOS
  //#if defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
    // Max 12-48 on MEGA
    #include <Servo.h>
    #define SERVOS 4
    int   servos_pin[SERVOS] = {2, 3, 4, 5};
    bool  servos_attached[SERVOS] = {false, false, false, false};
    int   servos_value[SERVOS] = {0, 0, 0, 0};
    Servo servos_Servo[SERVOS];
  //#endif
  
#endif


#ifdef USE_LEDRGB
  // Neopixels
  #ifndef LEDRGB_NUM
    #define LEDRGB_NUM 8+9+8+9
  #endif
  #ifndef LEDRGB_PIN
    #define LEDRGB_PIN 6
  #endif
  #include <Adafruit_NeoPixel.h>
  Adafruit_NeoPixel ledrgb_pixels = Adafruit_NeoPixel(LEDRGB_NUM, LEDRGB_PIN, NEO_GRB + NEO_KHZ800);
  //#define RAW_ASCII_START 32
  //#define RAW_ASCII_END RAW_ASCII_START+64
#endif



void sendPacket_volatile(byte cmd, byte seq) {
  byte check = 0xff ^ cmd ^ seq;
  SerialCommand.write(PACKET_MAGIC);
  SerialCommand.write(cmd);
  SerialCommand.write(seq);
  SerialCommand.write(check);
}

void sendPacket_safe(byte cmd, byte payloadLen, byte *payload) {
  // Construct packet
  
  // Increment sequence counter
  byte seq = (sPacketSeqLastSent + 1) & 0xff;

  // Calc checksum
  byte check = 0xff ^ cmd ^ seq ^ payloadLen;
  for (int i = 0; i < payloadLen; i++) {
    check ^= payload[i];
  }
  
  // Store for potential re-send
  sPacketCmdLast = cmd;
  sPacketPayloadLenLast = payloadLen;
  sPacketPayloadLast = payload;

  sPacketSeqLastSent = seq;
  
  // Actually transmit
  SerialCommand.write(PACKET_MAGIC);
  SerialCommand.write(cmd);
  SerialCommand.write(seq);
  SerialCommand.write(payloadLen);
  SerialCommand.write(payload, payloadLen);
  SerialCommand.write(check);
}


void sendCommand_data(byte cmd, byte len, byte *data) {
  sendPacket_safe(cmd, len, &data[0]);
}
void sendCommand_none(byte cmd) {
  // Just command, no payload
  sendPacket_safe(cmd, 0, NULL);
}
void sendCommand_byte(byte cmd, byte value) {
  // A command and one byte of data
  byte data[1] = {value};
  sendPacket_safe(cmd, 1, &data[0]);
}
void sendCommand_pin_byte(byte cmd, byte kind, byte index, byte value) {
  // A command and "pin parameters" using 1 byte value
  byte data[3] = {kind, index,  value };
  sendPacket_safe(cmd, 3, &data[0]);
}
void sendCommand_pin_word(byte cmd, byte kind, byte index, word value) {
  // A command and "pin parameters" using 1 word value
  byte data[4] = {kind, index,  byte(value >> 8), byte(value % 0x100) };
  sendPacket_safe(cmd, 4, &data[0]);
}



// Pin stuff

void handleCommand_pin_dir(byte kind, byte index, char dir) {
  switch(kind) {
    #ifdef USE_DIGITALS
    case KIND_DIGITAL:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Set dir DIGI #"); pyb_put_(index); pyb_put_(":"); pyb_put_(dir); pyb_put_(LF);
      #endif
      pinMode(digitals_pin[index], (dir == DIR_INPUT) ? INPUT : OUTPUT);
      break;
    #endif
    
    #ifdef USE_ANALOGS
    case KIND_ANALOG:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Set dir ANA #"); pyb_put_(index); pyb_put_(":"); pyb_put_(dir); pyb_put_(LF);
      #endif
      pinMode(analogs_pin[index], (dir == DIR_INPUT) ? INPUT : OUTPUT);
      break;
    #endif
  }
}

void handleCommand_pin_setDigital(byte kind, byte index, int value) {
  
  switch(kind) {
  
    #ifdef USE_DIGITALS
    case KIND_DIGITAL:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Set d value of DIGI #"); pyb_put_(index); pyb_put_(":"); pyb_put_(value); pyb_put_(LF);
      #endif
      digitalWrite(digitals_pin[index], value);
      break;
    #endif
    
    #ifdef USE_ANALOGS
    case KIND_ANALOG:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Set d value of ANA #"); pyb_put_(index); pyb_put_(":"); pyb_put_(value); pyb_put_(LF);
      #endif
      digitalWrite(analogs_pin[index], value);
      break;
    #endif
    
    #ifdef USE_SERVOS
    case KIND_SERVO:
    if (value == 0) {
        #ifdef DEBUG_VERBOSE
          pyb_put_("# Det. servo #"); pyb_put_(index); pyb_put_(LF);
        #endif
        servos_Servo[index].detach();
        servos_attached[index] = false;
      } else {
        #ifdef DEBUG_VERBOSE
          pyb_put_("# Att. servo #"); pyb_put_(servos_pin[index]); pyb_put_(LF);
        #endif
        servos_attached[index] = true;
        //servo->attach(servos_pin[index]);
        servos_Servo[index].attach(servos_pin[index]);
      }
      break;
    #endif
  }
  
}

void handleCommand_pin_setAnalog(byte kind, byte index, int value) {
  
  switch(kind) {
    #ifdef USE_DIGITALS
    case KIND_DIGITAL:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Set ana DIGI #"); pyb_put_(index); pyb_put_(" to "); pyb_put_(value); pyb_put_(LF);
      #endif
      analogWrite(digitals_pin[index], value);
      break;
    #endif
    
    #ifdef USE_ANALOGS
    case KIND_ANALOG:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Set ana ANA #"); pyb_put_(index); pyb_put_(" to "); pyb_put_(value); pyb_put_(LF);
      #endif
      // I don't think any Arduino supports PWM output from analog inputs...
      analogWrite(analogs_pin[index], value);
      break;
    #endif
    
    #ifdef USE_SERVOS
    case KIND_SERVO:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Set ana SERVO #"); pyb_put_(index); pyb_put_(" to "); pyb_put_(value); pyb_put_(LF);
      #endif
      
      //Servo *servo = &servos_Servo[index];
      if (servos_attached[index] == true) {
        servos_value[index] = value;
        //servo->write(value);
        servos_Servo[index].write(value);
      }
      break;
    #endif
  }
}

int handleCommand_pin_get(byte kind, byte index) {
  int value;
  
  switch(kind) {
    #ifdef USE_DIGITALS
    case KIND_DIGITAL:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Get DIGI #"); pyb_put_(index); pyb_put_(LF);
      #endif
      value = digitalRead(digitals_pin[index]);
      sendCommand_pin_byte(CMD_PUT, kind, index, value);
      break;
    #endif
    #ifdef USE_ANALOGS
    case KIND_ANALOG:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Get ANA #"); pyb_put_(index); pyb_put_(LF);
      #endif
      value = analogRead(analogs_pin[index]);
      sendCommand_pin_word(CMD_PUT, kind, index, value);
      break;
    #endif
    #ifdef USE_SERVOS
    case KIND_SERVO:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Get SERVO #"); pyb_put_(index); pyb_put_(LF);
      #endif
      value = servos_value[index];
      sendCommand_pin_word(CMD_PUT, kind, index, value);
      break;
    #endif
  }
  return value;
}

void handleCommand_pin_watch(byte kind, byte index, int thresh) {
  switch(kind) {
    #ifdef USE_DIGITALS
    case KIND_DIGITAL:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Watch DIGI #"); pyb_put_(index); pyb_put_(", th="); pyb_put_(thresh); pyb_put_(LF);
      #endif
      digitals_watch[index] = thresh;
      break;
    #endif
    
    #ifdef USE_ANALOGS
    case KIND_ANALOG:
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Watch ANA #"); pyb_put_(index); pyb_put_(", th="); pyb_put_(thresh); pyb_put_(LF);
      #endif
      analogs_watch[index] = thresh;
      break;
    #endif
  }
}


#ifdef USE_LEDRGB
void handleCommand_pin_ledRgb(byte kind, byte index, byte len, byte *data) {
  byte vR, vG, vB;
  int o;
  
  // Set RGB LED
  switch(kind) {
    case 'C':
      // Clear
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Clear LEDRGB\n");
      #endif
      for (int i = 0; i < LEDRGB_NUM; i++)
        ledrgb_pixels.setPixelColor(i, ledrgb_pixels.Color(0,0,0));

      // Do not automatically show. This allows fluid animations with little data exchange
      //ledrgb_pixels.show();
      break;
    
    case 'S':
      // Show contents
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Show LEDRGB\n");
      #endif
      ledrgb_pixels.show();
      break;

    case 'F':
      // Fill pixels
      vR = data[0]; //hex2int8(&s[3+COMMAND_DIGITS+1]);
      vG = data[1]; //hex2int8(&s[3+COMMAND_DIGITS+1+2]);
      vB = data[2]; //hex2int8(&s[3+COMMAND_DIGITS+1+2+2]);
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Fill RGB "); pyb_put_(index); pyb_put_(" to "); pyb_put_(vR); pyb_put_(","); pyb_put_(vG); pyb_put_(","); pyb_put_(vB); pyb_put_(LF);
      #endif
      if (index == 0) index = LEDRGB_NUM;
      for (int i = 0; i < index; i++) {
        ledrgb_pixels.setPixelColor(i, ledrgb_pixels.Color(vR, vG, vB));
      }
      break;

    case 'R':
      // Raw input
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Set RGB raw len="); pyb_put_(index); pyb_put_(LF);
      #endif
      o = 0;
      for (int i = 0; i < index; i++) {
        vR = data[o++]; //map((byte)s[o], RAW_ASCII_START, RAW_ASCII_END, 0, 255); o++;
        vG = data[o++]; //map((byte)s[o], RAW_ASCII_START, RAW_ASCII_END, 0, 255); o++;
        vB = data[o++]; //map((byte)s[o], RAW_ASCII_START, RAW_ASCII_END, 0, 255); o++;
        ledrgb_pixels.setPixelColor(i, ledrgb_pixels.Color(vR, vG, vB));
      }
      break;

    case 'P':
      // Set single pixel
      vR = data[0]; //hex2int8(&s[3+COMMAND_DIGITS+1]);
      vG = data[1]; //hex2int8(&s[3+COMMAND_DIGITS+1+2]);
      vB = data[2]; //hex2int8(&s[3+COMMAND_DIGITS+1+2+2]);
      #ifdef DEBUG_VERBOSE
        pyb_put_("# Set RGB pixel #"); pyb_put_(index); pyb_put_(" to "); pyb_put_(vR); pyb_put_(","); pyb_put_(vG); pyb_put_(","); pyb_put_(vB); pyb_put_(LF);
      #endif
      ledrgb_pixels.setPixelColor(index, ledrgb_pixels.Color(vR, vG, vB));
      break;

  }
  
}
#endif





void handleCommand_pin(byte cmd, byte kind, byte index, byte len, byte *data) {
  /*
  #ifdef DEBUG_VERBOSE
    pyb_put(F("Handle command..."));
  #endif
  */

  int value;
  char dir;
  int thresh;
  
  switch(cmd) {
    case CMD_DIR:
      // Set direction of a pin
      dir = data[0];
      handleCommand_pin_dir(kind, index, dir);
      break;
    
    case CMD_SET_DIGITAL:
      // Set (binary) value of a pin
      value = data[0];  //[hex2int8(&s[3+COMMAND_DIGITS+1]); //s.substring(3+COMMAND_DIGITS+1).toInt();
      handleCommand_pin_setDigital(kind, index, value);
      break;
    
    case CMD_SET_ANALOG:
      // Set value of a pin
      value = int(data[0]) << 8 + int(data[1]); //hex2int16(&s[3+COMMAND_DIGITS+1]); //s.substring(3+COMMAND_DIGITS+1).toInt();
      handleCommand_pin_setAnalog(kind, index, value);
      break;
    
    case CMD_GET:
      // Get value of a pin
      handleCommand_pin_get(kind, index);
      break;
      
    case CMD_WATCH:
      // Watch a pin
      //char w = s[3+COMMAND_DIGITS+1];
      thresh = (int(data[0]) << 8) + int(data[1]);  //s.substring(3+COMMAND_DIGITS+1).toInt();
      handleCommand_pin_watch(kind, index, thresh);
      break;
    
    #ifdef USE_LEDRGB
    case CMD_LEDRGB:
      handleCommand_pin_ledRgb(kind, index, len, data);
    #endif
    
  }
}



void handleCommand(byte cmd, byte len, byte *data) {

  // Handle "pin" commands
  switch(cmd) {
    case CMD_DIR:
    case CMD_SET_DIGITAL:
    case CMD_SET_ANALOG:
    case CMD_GET:
    case CMD_WATCH:
    case CMD_LEDRGB:
      if (len >= 2) {
        handleCommand_pin(cmd, data[0], data[1], len-2, &data[2]);
      } else {
        sendPacket_volatile(CMD_ERROR, ERROR_LENGTH);
      }
      return;
    
    // Handle other commands here      
    //case CMD_FOO:
    
    default:
      // Unknown command
      #ifdef DEBUG_VERBOSE
        pyb_put(F("Unknown cmd!"));
      #endif
      sendPacket_volatile(CMD_UNKNOWN, cmd);
  }

}




void handlePacket(byte cmd, byte seq, byte payloadLen, byte *payload, byte checkActual) {

  // Handle low level packet

  // Check integrity (checksum)
  //if (checkActual != checkSupposed) {
  if (checkActual != 0) {
    #ifdef DEBUG_VERBOSE
      pyb_put(F("Checksums differ!"));
    #endif
    sendPacket_volatile(CMD_NACK, seq);
    return;
  }

  
  // Handle low level commands
  int delta;
  byte sPacketSeqOld;
  
  switch (cmd) {
    case CMD_ACK:
      // Packet was received. Good.

      // Check if a number was skipped
      delta = (int)seq - (int)rPacketSeqLastAcked;
      // Handle wrap-around (e.g. lastAcked=0xff, nCmdSeq=0x02)
      if (delta < 0) delta += 0x100;
      
      if (delta != 1) {
        #ifdef DEBUG_VERBOSE
          pyb_put_(F("Delta != 1! lastAck="));
          pyb_put_(rPacketSeqLastAcked);
          pyb_put_(F(", nowAck="));
          pyb_put(seq);
        #endif
        // CMD_ERROR if differs more than 1
        //sendCommand_byte(CMD_ERROR, ERROR_SEQ_DELTA);
      }
      
      rPacketSeqLastAcked = seq;
      
      // Remove from waiting queue (if there is any)?
      
      return; // Don't ACK the ACK
      break;
    
    case CMD_NACK:
      // Packet was not received
    
      // See if we can re-send it...
      // Comapre given sequence number and our last known sequence number
      if (seq == sPacketSeqLastSent) {
        // It was the last packet sent. We could re-send it!
        #ifdef DEBUG_VERBOSE
          pyb_put(F("Re-send"));
        #endif
        
        // Re-send using sPacketCmdLast, sPacketSeqLast, sPacketPayloadLenLast, sPacketPayloadLast
        sPacketSeqOld = sPacketSeqLastSent; // Store old seq number
        
        sPacketSeqLastSent = seq-1;  // Set sequence number (minus one, because it will be incremented in the beginning)
        sendPacket_safe(sPacketCmdLast, sPacketPayloadLenLast, sPacketPayloadLast);  // Actually re-send
        
        sPacketSeqLastSent = sPacketSeqOld; // Restore seq number
        
      } else {
        #ifdef DEBUG_VERBOSE
          pyb_put(F("NACK seq obsolete!"));
        #endif
        sendPacket_volatile(CMD_ERROR, ERROR_OBSOLETE);
      }

      return; // Don't ACK the NACK
      break;

    case CMD_UNKNOWN:
    case CMD_ERROR:
      // Got error/unknown from host!?
      
      #ifdef DEBUG_VERBOSE
        pyb_put(F("CMD_UNKNOWN/CMD_ERROR"));
      #endif
      return; // Don't ACK the ERROR
      break;
    
  }
  
  // Ack the packet
  sendPacket_volatile(CMD_ACK, seq);
  

  #ifdef USE_CUSTOM_COMMANDS
    // Handle custom commands
    if (pybridge_on_command != NULL) {
      if (pybridge_on_command(cmd, payloadLen, &payload[0]))
        return; // Stop propagation if result > 0
    }
  #endif

  // Handle built-in commands
  handleCommand(cmd, payloadLen-1, &payload[1]);
}



void updateReadings() {
  // Refresh port readings
  int index;
  int value;
  
  #ifdef USE_DIGITALS
  // Refresh digitals
  for(index = 0; index < DIGITALS; index++) {
    if (digitals_watch[index] > 0) {
      value = digitalRead(digitals_pin[index]);
      if (value != digitals_value[index]) {
        sendCommand_pin_byte(CMD_PUT, KIND_DIGITAL, index, value);
        digitals_value[index] = value;
      }
    }
  }
  #endif
  
  #ifdef USE_ANALOGS
  // Refresh analogs
  for(index = 0; index < ANALOGS; index++) {
    if (analogs_watch[index] > 0) {
      value = analogRead(analogs_pin[index]);
      if (abs(value - analogs_value[index]) >= analogs_watch[index]) {
        sendCommand_pin_word(CMD_PUT, KIND_ANALOG, index, value);
        analogs_value[index] = value;
      }
      //analogs_value[index] = value;
    }
  }
  #endif
}


// Regularly check the serial port for new data
void checkCommands() {
  
  // Handle serial command input
  
  if (SerialCommand.available() == 0) {
    // No new data
    
    if (rPacketState == OFS_START) {
      // Nothing to do. Protocol is idling and there is no data.
      return;
    }
    
    // We are currently inside a command
    
    // Check if we have a timeout
    unsigned long nowMillis;
    int delayMillis;
    
    nowMillis = millis();
    delayMillis = nowMillis - rPacketStartMillis;
    if (delayMillis > COMMAND_TIMEOUT_MILLIS) {
      // Timeout has occured!
      
      #ifdef DEBUG_VERBOSE
        pyb_put(F("Timeout - Resetting protocol state"));
      #endif
      
      // Reset state
      rPacketState = OFS_START;
      
      // If we already have a sequence number: NACK it!
      if (rPacketState > OFS_SEQ) {
        #ifdef DEBUG_VERBOSE
          pyb_put(F("NACKing"));
        #endif
        sendCommand_byte(CMD_NACK, rPacketSeq);
      }
      
    }
    
    
  } else  // Serial available!
  while (SerialCommand.available()) {
    byte c = (byte)SerialCommand.read();

    // Checksum
    rPacketCheckActual ^= c;
    
    // Protocol handling...
    switch(rPacketState) {
      case OFS_START:
        if (c == PACKET_MAGIC) {
          // Record start of packet
          rPacketStartMillis = millis();
          
          // Reset packet memory
          rPacketCmd = 0;
          rPacketSeq = 0;
          rPacketPayloadLen = 0;
          rPacketCheckActual = 0xff;
          rPacketCheckSupposed = 0;
          
          rPacketState = OFS_CMD;
        } else {
          // Not magic: Ignore / discard
          #ifdef DEBUG_VERBOSE
            pyb_put(F("Not magic!"));
          #endif
        }
        break;

      case OFS_CMD:
        rPacketCmd = c;
        rPacketState = OFS_SEQ;
        break;
      
      case OFS_SEQ:
        rPacketSeq = c;

        // Short volatile packets without payload end here
        if (  (rPacketCmd == CMD_ACK)
           || (rPacketCmd == CMD_NACK)
           || (rPacketCmd == CMD_UNKNOWN)
           || (rPacketCmd == CMD_ERROR)
        ) {
            // No payload needed
            rPacketState = OFS_CHECK;  // Straight to checksum
        } else {
          // Continue with payload
          rPacketState = OFS_LEN;
        }
        break;
      
      case OFS_LEN:
        rPacketPayloadLen = c;
        //rPacketPayload = []
        rPacketPayloadLenCurrent = 0;
        if (rPacketPayloadLen > 0) rPacketState = OFS_DATA;
        else rPacketState = OFS_CHECK;
        break;
      
      case OFS_DATA:
        rPacketPayload[rPacketPayloadLenCurrent] = c;
        rPacketPayloadLenCurrent += 1;
        if (rPacketPayloadLenCurrent >= rPacketPayloadLen) {
          rPacketState = OFS_CHECK;
        }
        break;
      
      case OFS_CHECK:
        rPacketCheckSupposed = c;
        // Since the checksumActual is also XORed with this last byte, it should be zero if correct!
        handlePacket(rPacketCmd, rPacketSeq, rPacketPayloadLen, &rPacketPayload[0], rPacketCheckActual);
        rPacketState = OFS_START;
        break;
    }
    
  }
}



void pybridge_setup() {
  /*
  SerialCommand.begin(SERIAL_BAUD);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }
  */
  
  int i;
  
  // Init serial command
  //commandString.reserve(COMMAND_MAX_INPUT);
  rPacketState = OFS_START;
  rPacketStartMillis = 0;
  
  #ifdef USE_DIGITALS
    // Init digitals...
    for(i = 0; i < DIGITALS; i++) {
      digitals_value[i] = digitalRead(digitals_pin[i]);
    }
  #endif
  
  #ifdef USE_ANALOGS
    // Init analogs...
    for(i = 0; i < ANALOGS; i++) {
      analogs_value[i] = analogRead(analogs_pin[i]);
    }
  #endif
  
  #ifdef USE_SERVOS
    // Init servos
  #endif
  
  #ifdef USE_LEDRGB
    // Init LEDRGB Neopixels
    ledrgb_pixels.begin();
  #endif
  
  
  //pyb_put(F("Ready"));
  
  //sendCommand_none(CMD_READY);
}

void pybridge_loop() {
  checkCommands();
  updateReadings();
}
#endif

